"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Helper functions to swap instances of a type of family placed to another type,.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This helper function expect a folder containing swap directive files. For format of those files refer to module RevitFamilyRenameFilesUtils

Note:

- The revit category is used to uniquly identify the family to swap. It is assumed that the target family type belongs to a family of the same category.

"""


#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from duHast.Revit.Family.Utility import family_swap_instances_by_type_utils as rFamSwapUtils
from duHast.Revit.Family.family_functions import get_name_and_category_to_family_dict
from duHast.Revit.Family.family_utils import get_family_instances_by_symbol_type_id
from duHast.UI.Objects.ProgressBase import ProgressBase
from duHast.Revit.Common import transaction as rTran
from duHast.Utilities.Objects import result as res
from duHast.Revit.Family.Data.Objects.family_base_data_processor_defaults import (
    NESTING_SEPARATOR,
)

from Autodesk.Revit.DB import Element, Transaction


def get_target_type(doc, families, swap_directive):
    """
    Get the target type to swap to.

    :param doc: The current family document.
    :type doc: Autodesk.Revit.DB.Document
    :param families: dictionary of families where key is concatenated name + category and value is the family element.
    :type families: {str:Autodesk.Revit.DB.Family}
    :param swap_directive: The swap directive.
    :type swap_directive: class:`.FamilyDirectiveSwap`

    :return:
        Result class instance.

        - result.status. True if a single target family type was found, otherwise False.
        - result.message will contain each swap message
        - result.result list of target family types

        On exception:

        - result.status (bool) will be False.
        - result.message will contain an exception message in format: 'Failed to find target type for: ' + swap_directive.target_family_type_name
        - result.result will be empty
    
    :rtype: :class:`.Result`
    """

    return_value = res.Result()

    # get the family
    family_key = "{}{}{}".format(swap_directive.target_family_name,NESTING_SEPARATOR, swap_directive.category)

    # check if family is loaded
    if family_key not in families:
        return_value.update_sep(
            False, "Family not found: {}".format(family_key)
        )
        return return_value
    
    target_family = families[family_key]
            
    for fam_symbol_id in target_family.GetFamilySymbolIds():
        fam_symbol = doc.GetElement(fam_symbol_id)
        fam_symbol_name = Element.Name.GetValue(fam_symbol)
        if fam_symbol_name == swap_directive.target_family_type_name:
            return_value.append_message("...Found symbol: {}".format(fam_symbol_name))
            return_value.result.append(fam_symbol)
            return return_value
            
    return_value.update_sep(False, "...No target type found for: {}".format(swap_directive.target_family_type_name))
    return return_value


def get_super_component_id(doc, instance):

    """
    Get the super component id of a family instance.

    :param doc: The current family document.
    :type doc: Autodesk.Revit.DB.Document
    :param instance: The family instance.
    :type instance: Autodesk.Revit.DB.FamilyInstance

    :return:
        Result class instance.

        - result.status. True if a single super component was found, otherwise False.
        - result.message will contain each swap message
        - result.result list of super component ids (single entry)

        On exception:

        - result.status (bool) will be False.
        - result.message will contain an exception message in format: 'Failed to find super component for: ' + instance.Id
        - result.result will be empty
    
    :rtype: :class:`.Result`
    """

    return_value = res.Result()

    super_component = instance.SuperComponent

    if super_component == None:
        return_value.update_sep(False, "No super component found for: {}".format(instance.Id))
        # return an invalid element id integer
        return_value.result.append(-1)
        return return_value
    
    # get the super component id
    super_component_id = instance.SuperComponent.Id
    if super_component_id != None and super_component_id.IntegerValue > 0:
        return_value.append_message("...Found super component id: {}".format(super_component_id))
        # get the family of the super component
        super_fam_Id_int = super_component.Symbol.Family.Id.IntegerValue
        return_value.result.append(super_fam_Id_int)
    else:
        return_value.append_message("Failed to find super component for: {}".format(instance.Id))
        return_value.result.append(-1)

    return return_value


def get_group_id(doc, instance):
    """
    Get the group id of a family instance.

    :param doc: The current family document.
    :type doc: Autodesk.Revit.DB.Document
    :param instance: The family instance.
    :type instance: Autodesk.Revit.DB.FamilyInstance

    :return:
        Result class instance.

        - result.status. True if at least one group was found, otherwise False.
        - result.message will contain each swap message
        - result.result list of group ids

        On exception:

        - result.status (bool) will be False.
        - result.message will contain an exception message in format: 'Failed to find group for: ' + instance.Id
        - result.result will be empty
    
    :rtype: :class:`.Result`
    """

    return_value = res.Result()

    # get the group instance id
    group_id = instance.GroupId

    if group_id == None or group_id.IntegerValue == -1:
        return_value.update_sep(False, "No group found for: {}".format(instance.Id))
        # return an invalid element id integer
        return_value.result.append(-1)
        return return_value
    
    # get the group
    group = doc.GetElement(group_id)
    # get the group type id
    group_type_id_int = group.GroupType.Id.IntegerValue
    if group_type_id_int != None and group_type_id_int > 0:
        return_value.append_message("...Found group type id: {}".format(group_type_id_int))

        # check if there is more than 1 instance of this group
        group_sets = group.GroupType.Groups
        if group_sets.Size > 1:
            return_value.append_message("...Group is placed more than once.")
            return_value.result.append(group_type_id_int)
        else:
            # revit does allow change of group if only one instance is placed
            return_value.update_sep(False,"...Group is placed only once.")
            return_value.result.append(-1)
    else:
        return_value.append_message("Failed to find group for: {}".format(instance.Id))
        return_value.result.append(-1)

    return return_value


def _get_fam_instances(doc, family, swap_directive):
    """
    Get instances of a family type to be swapped.

    Note:
    Will return 3 lists: 
    - instances which can be swapped out.
    - instances which are nested in other families.
    - instances which are in groups, where the group is placed more than once.


    :param doc: The current family document.
    :type doc: Autodesk.Revit.DB.Document
    :param family: The family to get instances of.
    :type family: Autodesk.Revit.DB.Family
    :param swap_directive: The swap directive.
    :type swap_directive: class:`.FamilyDirectiveSwap`

    :return:
        Result class instance.

        - result.status. True if a single family type was found, otherwise False.
        - result.message will contain each swap message
        - result.result list of instances

        On exception:

        - result.status (bool) will be False.
        - result.message will contain an exception message in format: 'No instances found for type: ' + swap_directive.source_type_name
        - result.result will be empty
    
    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    
    # instances in model
    instances = []

    # a dictionary to hold count of instances by super component Family id integervalues
    super_components = {}

    # a dictionary to hold count of instances by group id integervalues
    groups = {}
    

    # get instances to be swapped
    for fam_symbol_id in family.GetFamilySymbolIds():
        fam_symbol = doc.GetElement(fam_symbol_id)
        fam_symbol_name = Element.Name.GetValue(fam_symbol)
        
        if fam_symbol_name ==  swap_directive.source_type_name :
            return_value.append_message("...Found symbol: {}".format(fam_symbol_name))
            # get all instances of this type
            instances_collector = get_family_instances_by_symbol_type_id(doc, fam_symbol.Id)
            for instance in instances_collector:
                
                # set default swap flag
                swap = True

                # check if instance is nested
                super_component_result = get_super_component_id(doc, instance)
                #return_value.update(super_component_result)

                # if this family is nested in another family do not swap
                if super_component_result.status:
                    # set swap flag to false
                    swap = False
                    super_component_id = super_component_result.result[0]
                    if super_component_id in super_components:
                        super_components[super_component_id] = super_components[super_component_id] + 1
                    else:
                        super_components[super_component_id] = 1
                
                # check if instance is in a group
                group_result = get_group_id(doc, instance)
                #return_value.update(group_result)

                # if this family is in a group do not swap
                if group_result.status:
                    # set swap flag to false
                    swap = False
                    group_type_id = group_result.result[0]
                    if group_type_id in groups:
                        groups[group_type_id] = groups[group_type_id] + 1
                    else:
                        groups[group_type_id] = 1
                
                # check if that instance can be swapped
                if swap:
                    instances.append(instance)

            return_value.append_message( "...Found {} instances of type: {}".format(len(instances), fam_symbol_name))
            break
    
    # update the return value as a tuple
    return_value.result.append((instances, super_components, groups))

    return return_value
    

def _swap_loaded_family_instances(doc, swap_directives, families, progress_callback=None):
    """
    Loops over nested families and if a match in rename directives is found will rename the family accordingly.

    :param doc: The current family document.
    :type doc: Autodesk.Revit.DB.Document
    :param rename_directives: List of rename directives.
    :type rename_directives: [rename_directive]
    :param family_ids: dictionary of families where key is concatenated name + category and value is the family element.
    :type family_ids: {str:Autodesk.Revit.DB.Family}

    :return:
        Result class instance.

        - result.status. True if all families where swapped successfully, otherwise False.
        - result.message will contain each rename messages in format 'Swapped family from :' +current Name + ' to ' + newName.
        - result.result empty list

        On exception:

        - result.status (bool) will be False.
        - result.message will contain an exception message in format: 'Failed to swap family from :' + currentName + ' to ' + newName
        - result.result will be empty

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    #return_value.status = False

    # progress call back
    callback_counter = 1

    for swap_directive in swap_directives:

        # update progress
        if progress_callback != None:
            progress_callback.update(callback_counter, len(swap_directives))

        # get the family
        family_key = "{}{}{}".format(swap_directive.name, NESTING_SEPARATOR, swap_directive.category)
        
        # check if family is loaded
        if family_key not in families:
            return_value.append_message(
                "Family has no swap directive: {}".format(family_key)
            )
            callback_counter = callback_counter + 1
            continue
        
        # revit family
        family = families[family_key]

        # instances in model to be swapped
        instances_result = _get_fam_instances(doc, family, swap_directive)
        
        # get instances to swap from returned tuple at index 0
        instances = instances_result.result[0][0]

        return_value.append_message("Found {} instances to swap out for family: {}".format(len(instances), family_key))

        if len(instances) == 0:
            # nothing to swap found move on
            callback_counter = callback_counter + 1
            continue

        # get the target type
        target_result = get_target_type(doc, families, swap_directive)
        
        if not target_result.status:
            # update return value with target result
            return_value.append_message("Family has no target type for: {}".format(swap_directive.target_family_type_name))
            callback_counter = callback_counter + 1
            # no target type found
            continue

        # get the target type to swap to
        target_type = target_result.result[0]
        return_value.append_message("Target type to swap to: {}".format(Element.Name.GetValue(target_type)))

        # set up an action swapping out instance
        def action():
            action_return_value = res.Result()
            try:
                # swap out the types
                for instance in instances:
                    instance.Symbol = target_type
                    return_value.append_message("......Swapped out type for instance: {}".format(instance.Id))
                action_return_value.update_sep(True, "Swapped out types successfully")
            except Exception as e:
                action_return_value.update_sep(False, "Failed to swap out types with exception: {}".format(e))
            return action_return_value
    
        tranny = Transaction(doc, "swapping out type: {} with {}".format(swap_directive.source_type_name, swap_directive.target_family_type_name))

        swap_result = rTran.in_transaction(tranny=tranny, action=action)
        return_value.update(swap_result)

        # update progress
        callback_counter = callback_counter + 1

        # check for progress cancel
        if progress_callback != None:
            if progress_callback.is_cancelled():
                return_value.append_message("User cancelled!")
                break
    
    return return_value


def swap_family_instances_of_types(doc, directory_path, progress_callback=None):
    """
    Entry point for this module. Will read swap directives files in given directory and attempt to swp out instances of types accordingly.

    Note: for swap directive file structure refer to module family_swap_instances_by_type_utils

    :param directory_path: Fully qualified directory path to where swap directive files are located.
    :type directory_path: str
    :return:
        Result class instance.

        - result.status. True if a single families was swapped successfully, otherwise False.
        - result.message will contain each swap message
        - result.result empty list

        On exception:

        - result.status (bool) will be False.
        - result.message will contain an exception message in format: 'Failed to swap family from :' + currentName + ' to ' + newName
        - result.result will be empty

    :rtype: :class:`.Result`
    """

    return_value = res.Result()

    # check callback class
    if progress_callback and isinstance(progress_callback, ProgressBase) == False:
        raise TypeError(
            "progress_callback needs to be inherited from ProgressBase. Got : {} instead.".format(
                type(progress_callback)
            )
        )
    
    # get directives from folder
    swap_directives_result = rFamSwapUtils.get_swap_directives(directory_path)
    # check if anything came back
    if swap_directives_result.status:
        # update return value
        return_value.update(swap_directives_result)
        swap_directives = swap_directives_result.result

        # get all family in file
        families = get_name_and_category_to_family_dict(doc)
        # check if any families are loaded
        if len(families) > 0:
            # swap instances as per directives
            swap_result= _swap_loaded_family_instances(
                doc=doc, swap_directives=swap_directives, families=families, progress_callback=progress_callback
            )
            # update return value
            return_value.update(swap_result)
        else:
            return_value.update_sep(True, "Mo loadable families in file.")
    else:
        return_value = swap_directives_result

    return return_value


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A number of functions around Revit independent tags elements tagged properties.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from Autodesk.Revit.DB import ElementId
from duHast.Revit.Common.revit_version import get_revit_version_number

HOST_ELEMENT_ID = "host_element_id"
LINK_ELEMENT_ID = "link_element_id"
LINK_INSTANCE_ID = "link_instance_id"


def get_tagged_elements(doc, tag):
    """
    Returns tagged element(s) depending on revit version.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param tag: The tag element
    :type tag: Autodesk.Revit.DB.IndependentTag

    :return: A list of dictionaries representing the tagged object(s).

        Dictionary has three keys:HOST_ELEMENT_ID,LINK_ELEMENT_ID,LINK_INSTANCE_ID
        If an exception occurs it will return a dictionary with same keys but None values.

    :rtype: [{str:Autodesk.Revit.DB.ElementId}]
    """

    # get the revit version:
    revit_version = get_revit_version_number(doc)
    data = None
    if revit_version <= 2021:
        data = get_tagged_elements_2021(tag)
    elif revit_version >= 2022:
        data = get_tagged_elements_2022(tag)
    return data


def get_tagged_elements_2021(tag):
    """
    Get tagged element Id in revit versions up to revit 2022

    :param tag: The tag element
    :type tag: Autodesk.Revit.DB.IndependentTag

    :return: A list with a single dictionary representing the tagged object.

        Dictionary has three keys:HOST_ELEMENT_ID,LINK_ELEMENT_ID,LINK_INSTANCE_ID
        If an exception occurs it will return a dictionary with same keys but None values.

    :rtype: [{str:Autodesk.Revit.DB.ElementId}]
    """

    data = []

    try:
        id = tag.TaggedElementId
        id_data = {}
        if id.HostElementId != ElementId.InvalidElementId:
            id_data[HOST_ELEMENT_ID] = id.HostElementId
            id_data[LINK_ELEMENT_ID] = ElementId.InvalidElementId
            id_data[LINK_INSTANCE_ID] = ElementId.InvalidElementId
        elif id.LinkedElementId != ElementId.InvalidElementId:
            id_data[HOST_ELEMENT_ID] = ElementId.InvalidElementId
            id_data[LINK_ELEMENT_ID] = id.LinkedElementId
            id_data[LINK_INSTANCE_ID] = ElementId.InvalidElementId
        else:
            id_data[HOST_ELEMENT_ID] = ElementId.InvalidElementId
            id_data[LINK_ELEMENT_ID] = ElementId.InvalidElementId
            id_data[LINK_INSTANCE_ID] = id.LinkInstanceId
        data.append(id_data)
    except:
        id_data = {}
        id_data[HOST_ELEMENT_ID] = None
        id_data[LINK_ELEMENT_ID] = None
        id_data[LINK_INSTANCE_ID] = None
        data.append(id_data)
    return data


def get_tagged_elements_2022(tag):
    """
    Get elbow properties in revit versions from revit 2023 onwards

    :param tag: The tag element
    :type tag: Autodesk.Revit.DB.IndependentTag

    :return: A list of dictionaries representing the tagged object(s).

        Dictionary has three keys:HOST_ELEMENT_ID,LINK_ELEMENT_ID,LINK_INSTANCE_ID
        If an exception occurs it will return a dictionary with same keys but None values.

    :rtype: [{str:Autodesk.Revit.DB.ElementId},]
    """

    data = []
    try:
        ids = tag.GetTaggedElementIds()
        for id in ids:
            id_data = {}
            try:
                if id.HostElementId != ElementId.InvalidElementId:
                    id_data[HOST_ELEMENT_ID] = id.HostElementId
                    id_data[LINK_ELEMENT_ID] = ElementId.InvalidElementId
                    id_data[LINK_INSTANCE_ID] = ElementId.InvalidElementId
                elif id.LinkedElementId != ElementId.InvalidElementId:
                    id_data[HOST_ELEMENT_ID] = ElementId.InvalidElementId
                    id_data[LINK_ELEMENT_ID] = id.LinkedElementId
                    id_data[LINK_INSTANCE_ID] = ElementId.InvalidElementId
                else:
                    id_data[HOST_ELEMENT_ID] = ElementId.InvalidElementId
                    id_data[LINK_ELEMENT_ID] = ElementId.InvalidElementId
                    id_data[LINK_INSTANCE_ID] = id.LinkInstanceId
            except:
                id_data[HOST_ELEMENT_ID] = None
                id_data[LINK_ELEMENT_ID] = None
                id_data[LINK_INSTANCE_ID] = None
            data.append(id_data)
    except:
        data.append(ElementId.InvalidElementId)
    return data



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to purging Revit walls. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


# required for .ToList() call
# import clr

# clr.AddReference("System.Core")
# from System import Linq

# clr.ImportExtensions(Linq)

from duHast.Revit.Common.purge_utils import get_used_unused_type_ids

from duHast.Revit.Walls.walls import (
    get_all_in_place_wall_type_ids,
    get_all_basic_wall_type_ids,
)
from duHast.Revit.Family.purge_unused_family_types import (
    get_unused_in_place_ids_for_purge,
)
from duHast.Revit.Walls.curtain_walls import get_all_curtain_wall_type_ids
from duHast.Revit.Walls.stacked_walls import get_all_stacked_wall_type_ids

# -------------------- used types --------------------------


def get_used_stacked_wall_type_ids(doc):
    """
    Returns all used stack wall type ids.
    Used: at least one instance of this type is placed in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing used stacked wall types.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = get_used_unused_type_ids(doc, get_all_stacked_wall_type_ids, 1)
    return ids


def get_used_in_place_wall_type_ids(doc):
    """
    Gets all used in place type ids in the model.
    Used: at least one instance of this type is placed in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing used in place wall types.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = get_used_unused_type_ids(doc, get_all_in_place_wall_type_ids, 1)
    return ids


def get_used_curtain_wall_type_ids(doc):
    """
    Gets type ids off all used curtain wall types.
    Used: at least one instance of this type is placed in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing used in curtain wall types.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = get_used_unused_type_ids(doc, get_all_curtain_wall_type_ids, 1)
    return ids


def get_used_basic_wall_type_ids(doc):
    """
    Gets type ids off all used basic wall types.
    Used: at least one instance of this type is placed in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing all basic wall types in use.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = get_used_unused_type_ids(doc, get_all_basic_wall_type_ids, 1)
    return ids


# -------------------- purge unused types --------------------------


def get_unused_basic_wall_type_ids_to_purge(doc):
    """
    Gets type ids off all unused basic wall types in model.
    This method can be used to safely delete unused basic wall types. In the case that no basic\
        wall instance using any of the types is placed, this will return all but one type id since\
        Revit requires at least one basic wall type definition to be in the model.
    
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing all basic wall types not in use.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = get_used_unused_type_ids(doc, get_all_basic_wall_type_ids, 0)
    # looks like a separate check is required whether any basic wall type is used in stacked wall type in model at this point
    # DOH! GetStackedWallMemberIds() is only available on wall element but not wallType. Why?
    available_type_count = len(get_all_basic_wall_type_ids(doc))
    if len(ids) == available_type_count:
        ids.pop(0)
    return ids


def get_unused_curtain_wall_type_ids_to_purge(doc):
    """
    Gets type ids off all unused curtain wall types.
    This method can be used to safely delete unused curtain wall types. In the case that no curtain\
        wall instance using any of the types is placed, this will return all but one type id since\
        Revit requires at least one curtain wall type definition to be in the model.
    
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing unused in curtain wall types.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = get_used_unused_type_ids(doc, get_all_curtain_wall_type_ids, 0)
    available_type_count = len(get_all_curtain_wall_type_ids(doc))
    if len(ids) == available_type_count:
        ids.pop(0)
    return ids


def get_unused_in_place_wall_type_ids(doc):
    """
    Gets all unused in place type ids in the model.
    Unused: Not one instance of this type is placed in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing unused in place wall types.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = get_used_unused_type_ids(doc, get_all_in_place_wall_type_ids, 0)
    return ids


def get_unused_in_place_wall_ids_for_purge(doc):
    """
    Gets symbol(type) ids and family ids (when no type is in use) of in place wall families which can be safely deleted from the model.
    This method can be used to safely delete unused in place wall types. There is no requirement by Revit to have at least one\
        in place wall definition in the model.
    
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing unused in place wall types and families.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = get_unused_in_place_ids_for_purge(doc, get_unused_in_place_wall_type_ids)
    return ids


def get_unused_stacked_wall_type_ids_to_purge(doc):
    """
    Gets all unused stacked wall type id's.
    This method can be used to safely delete unused wall types:
    In the case that no wall instance using any of the types is placed this will return all but one type id since\
        Revit requires at least one wall type definition to be in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing not used stacked wall types.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = get_used_unused_type_ids(doc, get_all_stacked_wall_type_ids, 0)
    available_type_count = len(get_all_stacked_wall_type_ids(doc))
    if len(ids) == available_type_count:
        ids.pop(0)
    return ids



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to purging Revit family types. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#
from duHast.Revit.Common.purge_utils import has_dependent_elements
from duHast.Utilities.utility import remove_items_from_list
from duHast.Revit.Family.family_utils import (
    get_family_symbols_ids,
    get_symbols_from_type,
)
from duHast.Revit.Family.Utility.loadable_family_categories import (
    CATEGORIES_LOADABLE_TAGS,
    CATEGORIES_LOADABLE_3D,
)


def family_all_types_in_use(fam_type_ids, used_type_ids):
    """
    Checks if symbols (types) of a family are in use in a model.
    Check is done by comparing entries of fam_type_ids with used_type_ids.
    :param fam_type_ids: list of symbol(type) ids of a family
    :type fam_type_ids: list Autodesk.Revit.DB.ElementId
    :param  used_type_ids: list of symbol(type) ids in use in a project
    :type used_type_ids: list Autodesk.Revit.DB.ElementId
    :return: False if a single symbol id contained in list fam_type_ids has a match in list used_type_ids, otherwise True.
    :rtype: bool
    """

    match = True
    for fam_type_id in fam_type_ids:
        if fam_type_id not in used_type_ids:
            match = False
            break
    return match


def get_unused_in_place_ids_for_purge(doc, unused_type_getter):
    """
    Filters symbol (type) IDs and family IDs of families.

    The returned list of IDs can be either unused family symbols or entire families if none of their symbols are in use.
    In terms of purging, it's faster to delete an entire family definition rather than deleting its symbols first and then the definition.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :param unused_type_getter: A function returning IDs of unused symbols (family types) as a list. It requires the current model document as an argument.
    :type unused_type_getter: function (doc) -> list of Autodesk.Revit.DB.ElementId

    :return: A list of Element IDs representing the family symbols and/or family IDs matching the filter.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    unused_ids = []
    unused_family_ids = []
    # get all unused type Ids
    unused_type_ids = unused_type_getter(doc)
    # get family Elements belonging to those type ids
    families = get_symbols_from_type(doc, unused_type_ids)
    # check whether an entire family can be purged and if so remove their symbol(type) ids from
    # from unusedType ids list since we will be purging the family instead
    for key, value in families.items():
        if family_all_types_in_use(value, unused_type_ids):
            unused_family_ids.append(key)
            unused_type_ids = remove_items_from_list(unused_type_ids, value)
    # check whether entire families can be purged and if so add their ids to list to be returned
    if len(unused_family_ids) > 0:
        unused_ids = unused_family_ids + unused_type_ids
    else:
        unused_ids = unused_type_ids
    return unused_ids


def get_used_unused_type_ids(doc, type_id_getter, use_type=0, exclude_shared_fam=True):
    """
    Filters types obtained by passing in the `type_id_getter` method and, depending on `use_type`, returns either the used or unused symbols of a family.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :param type_id_getter: A function returning IDs of symbols (family types) as a list. It requires the following arguments:
        - The current model document.
        - ICollection of built-in categories.
        - A boolean: exclude shared families.
    :type type_id_getter: function (doc, ICollection, bool) -> list of Autodesk.Revit.DB.ElementId

    :param use_type: 0 for no dependent elements (not used); 1 for has dependent elements (is in use).
    :type use_type: int

    :return: A list of Element IDs representing the family symbols matching the filter.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    # get all types elements available
    all_loadable_three_d_type_ids = type_id_getter(
        doc, CATEGORIES_LOADABLE_3D, exclude_shared_fam
    )
    all_loadable_tags_type_ids = type_id_getter(
        doc, CATEGORIES_LOADABLE_TAGS, exclude_shared_fam
    )
    all_type_ids = all_loadable_three_d_type_ids + all_loadable_tags_type_ids
    ids = []
    for type_id in all_type_ids:
        type = doc.GetElement(type_id)
        has_dependents = has_dependent_elements(doc, type)
        if has_dependents == use_type:
            ids.append(type_id)
    return ids


def get_unused_family_types(doc, exclude_shared_fam=True):
    """
    Filters unused non shared family (symbols) type ids in model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param exclude_shared_fam: Default is True (exclude any shared families from filter result)
    :type exclude_shared_fam: bool
    :return: A list of Element Ids representing the family symbols matching filter.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = get_used_unused_type_ids(doc, get_family_symbols_ids, 0, exclude_shared_fam)
    return ids


def get_unused_non_shared_family_symbols_and_type_ids_to_purge(doc):
    """
    Filters unused, non shared and in place family (symbols) type ids in model which can be purged from the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list of Element Ids representing the family symbols matching filter.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids_unused = get_unused_in_place_ids_for_purge(doc, get_unused_family_types)
    return ids_unused



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit ducts.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#
#

import Autodesk.Revit.DB as rdb
import Autodesk.Revit.DB.Mechanical as rdbM

from duHast.Revit.MEP_Systems.Utility.SymbolsInSystemTypes import (
    get_symbol_ids_of_mep_system_types,
    get_unique_ids_of_used_symbols_from_system_type_ids,
)
from duHast.Revit.MEP_Systems.Utility.RevitMEPSystemCategories import (
    CATS_LOADABLE_DUCTS,
)
from duHast.Revit.Common import common as com


def get_all_duct_types_by_category(doc):
    """
    Gets a filtered element collector of all duct types in the model.
    - round
    - oval
    - rectangular
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A filtered element collector of duct types
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = (
        rdb.FilteredElementCollector(doc)
        .OfCategory(rdb.BuiltInCategory.OST_DuctCurves)
        .WhereElementIsElementType()
    )
    return collector


def get_all_duct_types_by_class(doc):
    """
    Gets a filtered element collector of all duct types in the model.
    - round
    - oval
    - rectangular
    Will exclude in place families.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A filtered element collector of duct types
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    return rdb.FilteredElementCollector(doc).OfClass(rdbM.DuctType)


def get_all_duct_instances_in_model_by_category(doc):
    """
    Gets all duct elements placed in model.
    TODO: check these actually work...
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A filtered element collector of duct instances
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    return (
        rdb.FilteredElementCollector(doc)
        .OfCategory(rdb.BuiltInCategory.OST_DuctCurves)
        .WhereElementIsNotElementType()
    )


def get_all_duct_instances_in_model_by_class(doc):
    """
    Gets all duct elements placed in model.
    Will exclude in place families.
    TODO: check these actually work...
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A filtered element collector of duct instances
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    return (
        rdb.FilteredElementCollector(doc)
        .OfClass(rdbM.DuctType)
        .WhereElementIsNotElementType()
    )


def get_all_duct_type_ids_in_model_by_category(doc):
    """
    Gets all duct type ids available in model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list of ids representing duct types.
    :rtype: List Autodesk.Revit.DB.ElementId
    """

    ids = []
    col_cat = get_all_duct_types_by_category(doc)
    ids = com.get_ids_from_element_collector(col_cat)
    return ids


def get_all_duct_type_ids_in_model_by_class(doc):
    """
    Gets all duct type ids available in model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list of ids representing duct types.
    :rtype: List Autodesk.Revit.DB.ElementId
    """

    ids = []
    col_class = get_all_duct_types_by_class(doc)
    ids = com.get_ids_from_element_collector(col_class)
    return ids


# --------------------------------------- symbols used in MEP system types -------------------------------


def get_symbol_ids_used_in_duct_types(doc):
    """
    Gets a list of unique symbol ids used in system type properties of duct types.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of unique ids representing family symbols used in duct systems.
    :rtype: list  Autodesk.Revit.DB.ElementId
    """

    ids = []
    tye_ids = get_all_duct_type_ids_in_model_by_category(doc)
    ids = get_unique_ids_of_used_symbols_from_system_type_ids(doc, tye_ids)
    return ids


def get_symbol_ids_for_duct_types_in_model(doc):
    """
    Gets list of symbol ids of the following categories:
    - BuiltInCategory.OST_DuctAccessory,
    - BuiltInCategory.OST_DuctTerminal,
    - BuiltInCategory.OST_DuctFitting
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of ids representing family symbols.
    :rtype: list  Autodesk.Revit.DB.ElementId
    """

    ids = get_symbol_ids_of_mep_system_types(
        doc, CATS_LOADABLE_DUCTS, "GetSymbolIdsForDuctTypes"
    )
    return ids



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A number of helper functions relating to py pyrevit output to console duHast result class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from pyrevit import script
from duHast.Utilities.Objects.result import Result


def print_result(result, header=None):
    """
    Prints a result class instance to the pyRevit console.



    :param result: The result instance to be printed.
    :type result: Result
    :param header: The header to be printed before the result.
    :type header: str
    """

    # check if result is of type Result
    if not isinstance(result, Result):
        raise TypeError("result must be of type Result")

    # print header if it exists
    if header != None:
        print_header(header)

    # print the results message property
    # split the message by new line and print each line
    messages = result.message.split("\n")
    for m in messages:
        print("...{}".format(m))

    # print the status of the result
    print("Finished with status: {}!".format(result.status))


def print_header(header):
    """
    Prints a header to the pyRevit console.

    :param header: The header to be printed.
    :type header: str
    """

    output = script.get_output()
    if header != None:
        output.print_md("### {}".format(header))


def print_error(message):
    """
    Prints an error message to the pyRevit console.

    :param message: The error message to be printed.
    :type message: str
    """

    output = script.get_output()
    output.print_html('<div class="errorentry">{}</div>'.format(message))


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to CAD links.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import Autodesk.Revit.DB as rdb

from duHast.Revit.Common import delete as rDel, transaction as rTran
from duHast.Utilities.Objects import result as res
from duHast.Revit.Links.Utility.link_path import get_link_path


def get_all_cad_link_types(doc):
    """
    Gets all CAD link types in a model.
    Filters by class.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A filtered element collector of CAD link types
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = rdb.FilteredElementCollector(doc).OfClass(rdb.CADLinkType)
    return collector


def get_all_cad_link_instances(doc):
    """
    Gets all CAD link instances in a model.
    Filters by class.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A filtered element collector of import instances
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = rdb.FilteredElementCollector(doc).OfClass(rdb.ImportInstance)
    return collector


def get_cad_type_imports_only(doc):
    """
    Gets all CAD imports in a model.
    Filters by class and check whether the element is an external file reference (True its a link, False it is an import)
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list of all CAD imports in a model.
    :rtype: list Autodesk.Revit.DB.CADLinkType
    """

    cad_imports = []
    collector = rdb.FilteredElementCollector(doc).OfClass(rdb.CADLinkType)
    for cad in collector:
        if cad.IsExternalFileReference() == False:
            cad_imports.append(cad)
    return cad_imports


def sort_cad_link_types_by_model_or_view_specific(doc):
    """
    Returns two lists: First one: cad links types linked by view (2D) , second one cad link types linked into model (3D).
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: Two lists of cad link types.
    :rtype: list Autodesk.Revit.DB.CADLinkType, list Autodesk.Revit.DB.CADLinkType
    """

    cad_links_by_view = []
    cad_links_by_model = []
    collector_cad_types = get_all_cad_link_types(doc)
    collector_cad_instances = get_all_cad_link_instances(doc)
    ids_by_view = []
    # work out through the instance which cad link type is by view
    for c_instance in collector_cad_instances:
        if c_instance.ViewSpecific:
            ids_by_view.append(c_instance.GetTypeId())
    # filter all cad link types by id's identified
    for c_type in collector_cad_types:
        if c_type.Id in ids_by_view and c_type.IsExternalFileReference():
            cad_links_by_view.append(c_type)
        elif c_type.IsExternalFileReference():
            cad_links_by_model.append(c_type)
    return cad_links_by_view, cad_links_by_model


def get_all_cad_link_type_by_view_only(doc):
    """
    Gets all CAD links by view in a model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: Lists of cad link types.
    :rtype: list Autodesk.Revit.DB.CADLinkType
    """

    (
        cad_links_by_view,
        cad_links_by_model,
    ) = sort_cad_link_types_by_model_or_view_specific(doc)
    return cad_links_by_view


def get_all_cad_link_type_in_model_only(doc):
    """
    Gets all CAD links by model in a model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: Lists of cad link types.
    :rtype: list Autodesk.Revit.DB.CADLinkType
    """

    (
        cad_links_by_view,
        cad_links_by_model,
    ) = sort_cad_link_types_by_model_or_view_specific(doc)
    return cad_links_by_model


def delete_cad_links(doc):
    """
    Deletes all CAD links in a model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: Result class instance.

        - .status True if all CAD links got deleted. Otherwise False.
        - .message will contain status of deletion.

    :rtype: :class:`.Result`
    """

    ids = []
    return_value = res.Result()
    for p in rdb.FilteredElementCollector(doc).OfClass(rdb.ImportInstance):
        ids.append(p.Id)
    # delete all links at once
    return_value = rDel.delete_by_element_ids(
        doc, ids, "Deleting CAD links", "CAD link(s)"
    )
    return return_value


def reload_cad_links(
    doc, link_locations, host_name_formatted, do_something_with_link_name
):
    """
    Reloads CAD links from a given file location based on the original link type name (starts with comparison)

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param link_locations: A list of directories where CAD files can be located.
    :type link_locations: list str
    :param host_name_formatted: Not used yet
    :type host_name_formatted: TBC
    :param do_something_with_link_name: A function which amends the link name prior search for a match in folders.\
        I.e. can be used to truncate the link name i.e. the revision details of a link
    :type do_something_with_link_name: func(str) -> str
    :return: Result class instance.

        - .status True if all CAD links got reloaded successfully. Otherwise False.
        - .message will contain status of reload and fully qualified file name. On exception it will also include the exception message.
    
    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    try:
        # get all CAD link types in model
        for p in rdb.FilteredElementCollector(doc).OfClass(rdb.CADLinkType):
            link_type_name = do_something_with_link_name(rdb.Element.Name.GetValue(p))
            new_link_path = "unknown"
            try:
                new_link_path = get_link_path(link_type_name, link_locations, ".dwg")
                if new_link_path != None:
                    # reloading CAD links requires a transaction
                    def action():
                        action_return_value = res.Result()
                        try:
                            result = p.LoadFrom(new_link_path)
                            action_return_value.append_message (
                                link_type_name + " :: " + str(result.LoadResult)
                            )
                        except Exception as e:
                            action_return_value.update_sep(
                                False,
                                link_type_name
                                + " :: "
                                + "Failed with exception: "
                                + str(e),
                            )
                        return action_return_value

                    transaction = rdb.Transaction(doc, "Reloading: " + link_type_name)
                    reload_result = rTran.in_transaction(transaction, action)
                    return_value.update(reload_result)
                else:
                    return_value.update_sep(
                        False,
                        link_type_name
                        + " :: "
                        + "No link path or multiple path found in provided locations",
                    )
            except Exception as e:
                return_value.update_sep(
                    False, link_type_name + " :: " + "Failed with exception: " + str(e)
                )
    except Exception as e:
        return_value.update_sep(False, "Failed with exception: " + str(e))
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit category helper functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#


import clr
import System
from System.Collections.Generic import List

from duHast.Revit.Common.delete import delete_by_element_ids
from duHast.Utilities.Objects import result as res
from duHast.Revit.Common.transaction import in_transaction

from Autodesk.Revit.DB import BuiltInParameter, Transaction

#: subcategory renaming sampled dictionary
#: key is the current subcategory name, value is the new subcategory name
CAT_RENAMING = {"Clearance Zones": "AMAZING"}

#: list of built in parameters attached to family elements containing subcategory ids
ELEMENTS_PARAS_SUB = [
    BuiltInParameter.FAMILY_CURVE_GSTYLE_PLUS_INVISIBLE,
    BuiltInParameter.FAMILY_CURVE_GSTYLE_PLUS_INVISIBLE_MINUS_ANALYTICAL,
    BuiltInParameter.FAMILY_ELEM_SUBCATEGORY,
    BuiltInParameter.CLINE_SUBCATEGORY,
]


def get_main_sub_categories(doc):
    """
    Returns all subcategories of the family category in a dictionary where\
        key: sub category name
        value: sub category 

    :param doc: Current Revit family document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A dictionary.
    :rtype: dictionary {str: Autodesk.Revit.DB.Category}
    """

    cat_data = {}
    # get the family category
    family_category_name = doc.OwnerFamily.FamilyCategory.Name
    # get all subcategories in Document
    for main_cat in doc.Settings.Categories:
        # find the category matching this docs category
        # to ensure default subcategories with an id less then 0 are also extracted
        if main_cat.Name == family_category_name:
            # loop over all sub categories
            for sub_cat in main_cat.SubCategories:
                cat_data[sub_cat.Name] = sub_cat
    return cat_data


def get_available_categories_depending_on_category_type_owner(doc):
    """
    Gets all available categories from the document and filters it by the category type of the document.
    If the family category type is Model, then only model categories will be returned.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: List of categories filtered by the category type.
    :rtype: List[Category]
    """

    # get all available categories from the document
    categories = doc.Settings.Categories

    # use the categoryType of the current category to determine the category type to select from
    current_category_type = doc.OwnerFamily.FamilyCategory.CategoryType

    # filter the categories by the category type
    available_categories = [category for category in categories if category.CategoryType == current_category_type]

    return available_categories


def get_category_by_id(doc, category_id):
    """
    Gets the category from the document by its id. This includes any nested sub categories.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param category_id: Id of the category to get.
    :type category_id: ElementId

    :return: Category with the given id, or None if no category with the given id was found.
    :rtype: Category
    """

    # get the category from the document by its id
    categories = doc.Settings.Categories
    for cat in categories:
        if cat.Id == category_id:
            return cat
        # check any sub categories
        for sub_cat in cat.SubCategories:
            if sub_cat.Id == category_id:
                return sub_cat
           
    return None


def does_main_sub_category_exists(doc, sub_cat_name):
    """
    Checks whether a given subcategory exists in the family.

    Note: Only subcategory directly belonging to the family category will be checked for a match.

    :param doc: Current Revit family document.
    :type doc: Autodesk.Revit.DB.Document
    :param sub_cat_name: The name of the subcategory to be checked against.
    :type sub_cat_name: str

    :return: True if subcategory exists in family, otherwise False
    :rtype: bool
    """

    # get all sub categories belonging to family category
    sub_cats = get_main_sub_categories(doc)
    if sub_cat_name in sub_cats:
        return True
    else:
        return False


def delete_main_sub_category(doc, sub_cat_name):
    """
    Deletes a given subcategory from the family.

    Note: Only subcategory directly belonging to the family category will be checked for a match.

    :param doc: Current Revit family document.
    :type doc: Autodesk.Revit.DB.Document
    :param sub_cat_name: The name of the subcategory to be deleted.
    :type sub_cat_name: str

    :return: True if subcategory exists in family and was deleted successfully, otherwise False
    :rtype: bool
    """

    # get all sub categories belonging to family category
    sub_cats = get_main_sub_categories(doc)
    if sub_cat_name in sub_cats:
        # delete subcategory
        status_delete = delete_by_element_ids(
            doc,
            [sub_cats[sub_cat_name].Id],
            "delete subcategory: " + sub_cat_name,
            "subcategory",
        )
        return status_delete.status
    else:
        return False


def get_family_category(doc):
    """
    Gets the family category in a dictionary where\
        key: category name
        value: category 

    :param doc: Current Revit family document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A dictionary.
    :rtype: dictionary {str: Autodesk.Revit.DB.Category}
    """

    cat_data = {}
    # get the family category
    current_fam_cat = doc.OwnerFamily.FamilyCategory
    cat_data[current_fam_cat.Name] = current_fam_cat
    return cat_data


def get_other_sub_categories(doc):
    """
    Returns all family subcategories which do not belong to the actual family category.

    key: category name
    value: dictionary : key sub cat name, value: subcategory

    Note: custom subcategories have an Id greater 0

    :param doc: Current Revit family document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A dictionary.
    :rtype: dictionary {str: {str:Autodesk.Revit.DB.Category} }
    """

    cat_data = {}
    # get the family category
    family_category_name = doc.OwnerFamily.FamilyCategory.Name
    # get all subcategories in Document
    for main_cat in doc.Settings.Categories:
        # find the category not matching this docs category
        # to ensure default subcategories with an id less then 0 are also extracted
        if main_cat.Name != family_category_name:
            if main_cat.Name not in cat_data:
                cat_data[main_cat.Name] = {}
            # loop over all sub categories
            for sub_cat in main_cat.SubCategories:
                cat_data[main_cat.Name][sub_cat.Name] = sub_cat

    return cat_data


def get_other_custom_sub_categories(doc):
    """
    Returns all family custom subcategories which do not belong to the actual family category.
    Custom categories have an Id greater then 0.

    key: category name
    value: dictionary : key sub cat name, value: subcategory

    Note: custom subcategories have an Id greater 0

    :param doc: Current Revit family document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A dictionary.
    :rtype: dictionary {str: {str:Autodesk.Revit.DB.Category} }
    """

    cat_data = {}
    # get the family category
    family_category_name = doc.OwnerFamily.FamilyCategory.Name
    # get all subcategories in Document
    for main_cat in doc.Settings.Categories:
        # find the category not matching this docs category
        # to ensure default subcategories with an id less then 0 are also extracted
        if main_cat.Name != family_category_name:
            if main_cat.Name not in cat_data:
                cat_data[main_cat.Name] = {}
            # loop over all sub categories
            for sub_cat in main_cat.SubCategories:
                if sub_cat.Id.IntegerValue > 0:
                    cat_data[main_cat.Name][sub_cat.Name] = sub_cat
    return cat_data


def get_other_categories(doc):
    """
    Returns all family pre defined categories which do not belong to the actual family category.

    :param doc: Current Revit family document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A list of categories.
    :rtype: [Autodesk.Revit.DB.Category]
    """

    cat_data = []
    # get the family category
    family_category_name = doc.OwnerFamily.FamilyCategory.Name
    # get all subcategories in Document
    for main_cat in doc.Settings.Categories:
        # find the category not matching this docs category
        # to ensure default subcategories with an id less then 0 are also extracted
        if main_cat.Name != family_category_name:
            if main_cat not in cat_data:
                cat_data.append(main_cat)
    return cat_data


def get_category_by_built_in_def_name(doc, built_in_defs):
    """
    Returns categories by their built in definition

    :param doc: Current Revit family document.
    :type doc: Autodesk.Revit.DB.Document
    :param built_in_defs: list of BuiltInCategory Enumeration values
    :type built_in_defs: [Autodesk.Revit.DB.BuiltInCategory]

    :return: list of categories
    :rtype: [Autodesk.Revit.DB.Category]
    """

    cats = []
    document_settings = doc.Settings
    groups = document_settings.Categories
    for built_in_def in built_in_defs:
        cat = groups.get_Item(built_in_def)
        if cat != None:
            cats.append(cat)
    return cats


def set_family_category(doc, new_category_name):
    """
    Changes the family category to new one specified by name. (this will not re-instate any custom sub categories created under the new family category)

    :param doc: Current Revit family document.
    :type doc: Autodesk.Revit.DB.Document
    :param new_category_name: The name of the new family category.
    :type new_category_name: str

    :return: True only if the category was changed successfully. Any other case False! (That includes situations when the family is already of the new category)
    :rtype: bool
    """

    return_value = res.Result()
    cat = doc.OwnerFamily.FamilyCategory
    if cat.Name != new_category_name:
        if doc.Settings.Categories.Contains(new_category_name):

            def action():
                doc.OwnerFamily.FamilyCategory = doc.Settings.Categories.get_Item(
                    new_category_name
                )

            transaction = Transaction(
                doc, "Changing family category to:" + str(new_category_name)
            )
            change_cat = in_transaction(transaction, action)
            if change_cat.status:
                return_value.update_sep(
                    True,
                    "Successfully changed family category to: {}".format(
                        new_category_name
                    ),
                )
            else:
                return_value.update(change_cat)
        else:
            return_value.update_sep(
                False, "Invalid Category name supplied: {}".format(new_category_name)
            )
    else:
        return_value.update_sep(
            False, "Family is already of category: {}".format(new_category_name)
        )
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Utility functions writing / reading json objects to/ from file.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import traceback

from duHast.Utilities.Objects import result as res
from duHast.Utilities.files_get import get_files_single_directory

import codecs
import json
import os


def serialize(obj):
    """
    Serialize the object for JSON output, using to_json() if available.

    :param obj: The object to serialize.
    :return: A dictionary representation of the object.
    """
    if hasattr(obj, "class_to_dict") and callable(getattr(obj, "class_to_dict")):
        return json.loads(
            json.dumps(obj.class_to_dict(), indent=None)
        )  # Use the class_to_dict method
    else:
        return obj.__dict__  # Fallback to default


def serialize_utf(obj):
    """
    Serialize the object for JSON output including utf 8, using to_json_utf() if available.

    :param obj: The object to serialize.
    :return: A dictionary representation of the object.
    """
    if hasattr(obj, "to_json_utf") and callable(getattr(obj, "to_json_utf")):
        return json.loads(obj.to_json_utf())  # Use the to_json_utf method
    else:
        return obj.__dict__  # Fallback to default


def write_json_to_file(json_data, data_output_file_path, enforce_utf8=True):
    """
    Writes collected data to a new JSON formatted file.

    :param json_data: A dictionary to be written to file.
    :param data_output_file_path: Fully qualified file path to JSON data file.
    :param enforce_utf8: Will encode any string value as UTF-8, Default is True (recommended!!).

    :return:
        Result class instance.

        - result.status (bool) True if file was written without an exception, otherwise False.
        - result.message contains log messages.
        - result.result will be a single entry list containing json data as string.

        On exception:

        - result.status (bool) will be False.
        - result.message will contain exception message.
         result.result be an empty list
    :rtype: :class:`.Result`
    """

    result = res.Result()
    # file placeholder
    f = None
    try:

        json_string = None
        # Check if UTF-8 is to be enforced
        if enforce_utf8:

            json_string = json.dumps(
                json_data, indent=None, default=serialize_utf, ensure_ascii=False
            )
            # write data with codecs to ensure utf-8 encoding (slow)
            with codecs.open(data_output_file_path, "w", encoding="utf-8") as f:
                f.write(json_string)

        else:
            json_string = json.dumps(json_data, indent=None, default=serialize)

            # write data without codecs (fast)?
            with open(data_output_file_path, "w") as f:
                f.write(json_string)

        result.update_sep(
            True, "Data written to file: {}".format(data_output_file_path)
        )

        result.result.append(json_string)
    except Exception as e:
        tb = traceback.format_exc().strip().split("\n")
        result.update_sep(
            False,
            "Failed to write data to file with exception: {}. Trace back: {}".format(
                e, "::".join(tb)
            ),
        )
    finally:
        # make sure to close the file
        if f is not None:
            f.close()
    return result


def read_json_data_from_file(file_path):
    """
    Reads json from file in utf-8 encoded format.

    :param file_path: Fully qualified file path of json file to read.
    :type file_path: str

    :return:
        Result class instance.

        - result.status (bool) True if file was read without an exception, otherwise False.
        - result.message contains log messages.
        - result.result will contain a dictionary with the data read from the file.

        On exception:

        - result.status (bool) will be False.
        - result.message will contain exception message.
         result.result be an empty list
    :rtype: :class:`.Result`
    """

    result = res.Result()
    data = {}
    try:
        # Opening JSON file as utf-8
        with codecs.open(file_path, "r", encoding="utf-8") as f:
            # returns JSON object as
            # a dictionary
            data = json.load(f)
            f.close()

        # store the data in the result object
        result.result.append(data)
        result.update_sep(True, "Data read from file: {}".format(file_path))

    except Exception as e:
        result.update_sep(
            False, "Failed to read data to file with exception: {}".format(e)
        )
    return result


def combine_files_json(
    folder_path,
    file_prefix="",
    file_suffix="",
    file_extension=".txt",
    output_file_name="result.txt",
    file_getter=get_files_single_directory,
):
    """
    Combines multiple json formatted text files into a single json list formatted file, where each file is a list entry.
    Assumes:

    - each file can contain a single line json formatted string

    The new file will be saved into the same folder as the original files.

    :param folder_path: Folder path from which to get files to be combined and to which the combined file will be saved.
    :type folder_path: str
    :param file_prefix: Filter: File name starts with this value
    :type file_prefix: str
    :param file_suffix: Filter: File name ends with this value.
    :type file_suffix: str
    :param file_extension: Filter: File needs to have this file extension
    :type file_extension: str, format '.extension'
    :param out_put_file_name: The file name of the combined file, defaults to 'result.txt'
    :type out_put_file_name: str, optional
    :param file_getter: Function returning list of files to be combined, defaults to GetFilesSingleFolder
    :type file_getter: func(folder_path, file_prefix, file_suffix, file_extension), optional

    :return:
        Result class instance.

        - result.status (bool) True if file where combined without an exception, otherwise False.
        - result.message contains log messages.
        - result.result will contain a dictionaries with the data read from the file(s).

        On exception:

        - result.status (bool) will be False.
        - result.message will contain exception message.
         result.result be an empty list
    :rtype: :class:`.Result`

    """

    result = res.Result()
    try:
        # get all files to be combined
        file_list = file_getter(folder_path, file_prefix, file_suffix, file_extension)

        # read json data into a list of json objects
        json_objects = []
        for file in file_list:
            read_result = read_json_data_from_file(file_path=file)
            result.update(read_result)
            if read_result.status == False:
                return result
            json_objects.append(read_result.result[0])

        # write json data out
        result_write = write_json_to_file(
            json_data=json_objects,
            data_output_file_path=os.path.join(folder_path, output_file_name),
        )
        result.update(result_write)
    except Exception as e:
        result.update_sep(False, "Failed to combine files with exception: {}".format(e))

    return result



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Helper functions to read swap instances by type  directives file(s) and return them as a tuple to the caller
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

These helper function expect a text file in csv format with 5 columns:

- SOURCE_FAMILY_NAME = 0
- SOURCE_FAMILY_CATEGORY_NAME = 1
- SOURCE_FAMILY_TYPE_NAME = 2
- TARGET_FAMILY_NAME = 3
- TARGET_FAMILY_TYPE = 4

Note:

- First row is treated as a header row and its content is ignored.

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities import files_csv as fileCSV, files_get as fileGet
from duHast.Utilities.Objects import result as res
from duHast.Revit.Family.Data.Objects.family_directive_swap_instances_of_type import FamilyDirectiveSwap
from duHast.Utilities.files_io import file_exist




def _read_swap_directives(files):
    """
    Reads list of swap directives from file into named directive class instances.

    :param filePath: Fully qualified file path to rename directives file.
    :type filePath: str
    :return: List of swap directives.
    :rtype: [FamilyDirectiveSwap]
    """

    rename_directives = []
    for file in files:
        rows_result = fileCSV.read_csv_file(file)
        # check whether file was read successfully
        if rows_result.status is False:
            return rename_directives
        
        rows = rows_result.result
        
        # read rows in tuples ignoring the header row
        for i in range(1, len(rows)):
            if len(rows[i]) >= 4:

                data = FamilyDirectiveSwap(
                    name=rows[i][
                        FamilyDirectiveSwap.SWAP_DIRECTIVE_LIST_INDEX_SOURCE_FAMILY_NAME
                    ],
                    category=rows[i][
                        FamilyDirectiveSwap.SWAP_DIRECTIVE_LIST_INDEX_SOURCE_FAMILY_CATEGORY_NAME
                    ],
                    source_type_name=rows[i][
                        FamilyDirectiveSwap.SWAP_DIRECTIVE_LIST_INDEX_SOURCE_FAMILY_TYPE_NAME
                    ],
                    target_family_name=rows[i][
                        FamilyDirectiveSwap.SWAP_DIRECTIVE_LIST_TARGET_FAMILY_NAME
                    ],
                    target_family_type_name=rows[i][
                        FamilyDirectiveSwap.SWAP_DIRECTIVE_LIST_TARGET_FAMILY_TYPE
                    ],
                )

            rename_directives.append(data)
    return rename_directives

def get_swap_directives(directory_path):
    """
    Retrieves file swap directives from a given folder location.

    :param directory_path: Fully qualified folder path to folder containing directives.
    :type directory_path: str

    :return:
        Result class instance.

        - result.status. True if rename directives where found and loaded successfully, otherwise False.
        - result.message will contain number of directives found in format:'Found rename directives: ' + number
        - result.result list of directives

        On exception:

        - result.status (bool) will be False.
        - result.message will contain an exception message.
        - result.result will be empty

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    # check whether csv files matching file name filter exist in directory path
    swap_directive_files = fileGet.get_files_from_directory_walker_with_filters(
        directory_path,
        FamilyDirectiveSwap.SWAP_DIRECTIVE_FILE_NAME_PREFIX,
        "",
        FamilyDirectiveSwap.SWAP_DIRECTIVE_FILE_EXTENSION,
    )

    # check whether any files where found?
    if len(swap_directive_files) > 0:
        # attempt to re swap directives from files
        swap_directives = _read_swap_directives(swap_directive_files)
        # check whether any swap directives where found in files
        if len(swap_directives) > 0:
            return_value.update_sep(
                True, "Found swap directives: {}".format(len(swap_directives))
            )
            # store swap directives in result object
            return_value.result = swap_directives
        else:
            return_value.update_sep(
                False,FamilyDirectiveSwap.EXCEPTION_EMPTY_SWAP_DIRECTIVE_FILES
            )
    else:
        return_value.update_sep(
            False, FamilyDirectiveSwap.EXCEPTION_NO_SWAP_DIRECTIVE_FILES
        )

    return return_value


def write_swap_directives_to_file(swap_directives, file_path):
    """
    Writes swap directives to a specified file in CSV format.

    :param swap_directives: List of swap directives to write to file.
    :type swap_directives: list of FamilyDirectiveSwap
    :param file_path: Fully qualified file path where the directives will be written.
    :type file_path: str
    :return: Result object indicating success or failure of the write operation.
    :rtype: Result
    """
    
    return_value = res.Result()


    # loop over directives and convert them to a list of lists
    swap_directives_list = []
    try:
        for directive in swap_directives:
            swap_directives_list.append([
                directive.name,
                directive.category,
                directive.source_type_name,
                directive.target_family_name,
                directive.target_family_type_name
            ])

        # set default write mode and header
        write_mode = "w"
        header = FamilyDirectiveSwap.SWAP_DIRECTIVE_HEADER_ROW

        # check whether file exists and set write mode and header accordingly
        if file_exist(file_path):
            # file does not exist, so write mode is 'w'
            write_mode = "a"
            header =[]
        
        # write the directives to the file
        return_value = fileCSV.write_report_data_as_csv(
            file_name=file_path,
            header = header,
            write_type=write_mode,
            data=swap_directives_list,
            quoting=fileCSV.csv.QUOTE_MINIMAL,
        )

    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to write swap directives with exception: {}".format(e),
        )
    
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A matrix class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Supports matrices from 1 x 1 up to 4 x 4.


"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import json
import copy

from duHast.Utilities.Objects.base import Base
from duHast.Geometry.geometry_property_names import GeometryPropertyNames
from duHast.Geometry.Exceptions.incompatible_matrix_dimension import (
    IncompatibleMatrixDimensions,
)
from duHast.Utilities.compare import is_close


class Matrix(Base):
    def __init__(self, rows=None, cols=None, elements=None, j=None):
        """
        A basic matrix class. Matrices up to size of 4x4 are supported only.

        :param rows: Number of rows, defaults to None
        :type rows: int, optional
        :param cols: Number of columns, defaults to None
        :type cols: int, optional
        :param elements: The matrix, defaults to None
        :type elements: [[float]], optional
        :param j: A json formatted settings string or dictionary, defaults to None
        :type j: str or dic, optional
        :raises TypeError: Thrown if row is not an integer and no json is provided
        :raises TypeError: Thrown if cols is not an integer and no json is provided
        :raises ValueError: Thrown if matrix size exceeds 4 x 4 or is rows or columns is less than 1
        :raises ValueError: Thrown if elements provided do not match rows and cols count
        """

        super(Matrix, self).__init__()

        if j is not None:
            self._init_from_json(j)
        else:
            if not isinstance(rows, int):
                raise TypeError(
                    "rows must be of type int, got {} instead.".format(type(rows))
                )
            if not isinstance(cols, int):
                raise TypeError(
                    "cols must be of type int, got {} instead.".format(type(cols))
                )

            if rows < 1 or cols < 1 or rows > 4 or cols > 4:
                raise ValueError("Matrix dimensions must be between 1 and 4.")

            self._rows = rows
            self._columns = cols

            if elements is None:
                self._data = [
                    [0.0 for _ in range(self._columns)] for _ in range(self._rows)
                ]
            else:
                # make sure elements are of the right type and size
                self._validate_elements(elements)

    def _init_from_json(self, json_string):
        """
        Initialise class using the past in json string.

        :param json_string: A json formatted settings string or dictionary.
        :type json_string: str or dic
        :raises ValueError: Input must be a JSON string or a dictionary
        """

        try:
            if isinstance(json_string, str):
                json_string = json.loads(json_string)
            elif not isinstance(json_string, dict):
                raise TypeError("Input must be a JSON string or a dictionary.")

            self._rows = json_string.get(GeometryPropertyNames.ROWS, 0)
            self._columns = json_string.get(GeometryPropertyNames.COLUMNS, 0)

            elements_from_json = json_string.get(
                GeometryPropertyNames.DATA,
                [[0.0 for _ in range(self._columns)] for _ in range(self._rows)],
            )
            self._validate_elements(
                elements_from_json
            )  # Validate after loading from JSON
        except Exception as e:
            raise ValueError("Invalid JSON input: {}".format(e))

    def _validate_elements(self, elements):
        """Validate that the elements are all floats."""
        if len(elements) != self._rows or any(
            len(row) != self._columns for row in elements
        ):
            raise ValueError("Elements must match the specified dimensions.")

        for row in elements:
            for value in row:
                if not isinstance(value, (float, int)):  # Allow both float and int
                    raise TypeError("All elements must be of type float or int.")

        self._data = elements

    @property
    def rows(self):
        """Read-only property for the number of rows."""
        return self._rows

    @property
    def columns(self):
        """Read-only property for the number of columns."""
        return self._columns

    @property
    def data(self):
        """Read-only property for the matrix data."""
        return copy.deepcopy(self._data)

    def __getitem__(self, idx):
        return self._data[idx]

    def __setitem__(self, idx, value):
        if len(value) != self.columns:
            raise ValueError("Row must have exactly {} elements.".format(self.columns))
        self._data[idx] = value

    def __add__(self, other):
        if not isinstance(other, Matrix):
            raise TypeError(
                "other must be of type matrix, got {} instead.".format(other)
            )
        elif self.rows != other.rows or self.columns != other.columns:
            raise IncompatibleMatrixDimensions(
                "Can only add another matrix with the same dimensions.", other
            )
        return Matrix(
            self.rows,
            self.columns,
            [
                [self._data[i][j] + other[i][j] for j in range(self.columns)]
                for i in range(self.rows)
            ],
        )

    def __str__(self):
        return "\n".join(["\t".join(map(str, row)) for row in self._data])

    def __eq__(self, other):
        if not isinstance(other, Matrix):
            return NotImplemented

        if self.rows != other.rows or self.columns != other.columns:
            return False

        return all(
            is_close(self._data[i][j], other[i][j])
            for i in range(self.rows)
            for j in range(self.columns)
        )

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        # Convert self._data to a tuple of tuples for hashing
        data_as_tuple = tuple(tuple(row) for row in self._data)
        return hash((self.rows, self.columns, data_as_tuple))


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A class used to implement deleted element and modified element count for room separation lines in warnings solver.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Modify delete

- checks whether the deleted element count is a list of 2 entries
- checks whether any entry is a model line

Modified modifier

- check if any room in that collection has an area of zero after modification

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import System
# required since we are dealing with a c# List element
from System.Collections.Generic import  List

from duHast.Revit.Purge.Objects.ModifierBase import ModifierBase

# required for isInstance check
from Autodesk.Revit.DB import (
    CurveElement, 
    ElementId,
    FilteredElementCollector
)

from Autodesk.Revit.DB.Architecture import Room

class RoomSeparationLinesPurgeModifier(ModifierBase):
    def __init__(self, doc):
        """
        Class constructor.

        """

        super(RoomSeparationLinesPurgeModifier, self).__init__()

        # get all model lines in the project before deletion in order to be able to check if the deleted element is a model line
        self.model_lines = FilteredElementCollector(doc).OfClass(CurveElement ).ToElementIds()

        self.debug_log = []


    def  modify_modified(self, doc, modified):
        """
        Base implementation override to modify the modified element count.

        Returns modified element list unchanged if:
        - modified is not a c# List[ElementId]
        

        Returns:
        - modified element list without any room objects

        Args:
            modified: The modified element count

        """

        if isinstance(modified, List[ElementId]) == False:
            self.debug_log.append(
                "modified is not a List[ElementId]: {}".format(modified)
            )
            raise TypeError("deleted must be a List[ElementId]:".format(modified))

        
        # remove any other modellines from modified list
        # check if room is in modified list and if so if it has an area???
        filtered_ids = []

        self.debug_log.append(
            "...in rooms modified modifier: modified ids count: {}".format(len(modified))
        )

        for id in modified:
            element = doc.GetElement(id)
            if type(element) is Room:
                # if the room has no area, the change of that room separation line is important and should not go ahead
                self.debug_log.append(
                    "...in rooms modified modifier: Room found with area: {}".format(element.Area)
                )
                if element.Area > 0 is False:
                    filtered_ids.append(id)
            
            else:
                # all the other elements:
                # ModelLines and other elements can be ignored
                self.debug_log.append("...in rooms modified modifier: other element found: {}".format(type(element)))
                pass
        
        self.debug_log.append(
            "...in rooms modified modifier: filtered ids count: {}".format(len(filtered_ids))
        )
        return filtered_ids
    

    def  modify_deleted(self, doc, deleted):
        """
        Base implementation override to modify the deleted element count.

        Returns deleted element list unchanged if:
        - deleted is not a c# List[ElementId]
        - deleted is not a list of 2 elements

        Returns:
        - deleted element list without any model lines

        :param deleted: The deleted element id list
        :type deleted: List[ElementId]
        """

        if isinstance(deleted, List[ElementId]) == False:
            self.debug_log.append(
                "modified is not a List[ElementId]: {}".format(deleted)
            )
            raise TypeError("deleted must be a List[ElementId]:".format(deleted))
        
        # check if 2 elements got deleted..one is a model line and the other is <other> object
        # if so, remove the model line from the list

        if len(deleted) != 2:
            return deleted
        
        # check if the element is a model line
        filtered_ids = []
        for id in deleted:
            try:
                if id not in self.model_lines:
                    self.debug_log.append("...in rooms deleted modifier: non model line found: id: {} type:{}".format(id, type(doc.GetElement(id))))
                    filtered_ids.append(id)
            except Exception as e:
                self.debug_log.append("in deleted modifier: element not found: {} for : {}".format(e, id))
                print("in deleted modifier: element not found: {} for : {}".format(e, id))
        
        self.debug_log.append("...in rooms deleted modifier: filtered ids count: {}".format(len(filtered_ids)))
        return filtered_ids
        


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Helper functions relating to file IO operations. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import os
import shutil
import re

# from System.IO import Path


def file_exist(full_file_path):
    """
    Checks whether a file exists
    :param full_file_path: Fully qualified file path
    :type full_file_path: str
    :return: True file exists, otherwise False
    :rtype: bool
    """

    try:
        value = os.path.isfile(full_file_path)
    except Exception:
        value = False
    return value


def file_delete(full_file_path):
    """
    Deletes file.

    :param full_file_path: Fully qualified file path
    :type full_file_path: str
    :return: True if deleted, otherwise False
    :rtype: bool
    """
    try:
        os.remove(full_file_path)
        value = True
    except Exception:
        value = False
    return value


def get_directory_path_from_file_path(file_path):
    """
    Extracts directory from file path.

    :param file_path: A fully qualified file path.
    :type file_path: str
    :return: If no exception occurs : A fully qualified directory path,else an empty string.
    :rtype: str
    """
    try:
        value = os.path.dirname(file_path)
    except Exception:
        value = ""
    return value


def rename_file(old_name, new_name):
    """
    Renames a file.
    :param old_name: Fully qualified file path to file to be renamed.
    :type old_name: str
    :param new_name: Fully qualified new file name.
    :type new_name: str
    :return: True file renamed, otherwise False
    :rtype: bool
    """

    try:
        os.rename(old_name, new_name)
        value = True
    except Exception:
        value = False
    return value


def copy_file(old_name, new_name):
    """
    Copies a file
    :param old_name: Fully qualified file path to file to be copied.
    :type old_name: str
    :param new_name: Fully qualified path to new file location and name.
    :type new_name: str
    :return: True file copied, otherwise False
    :rtype: bool
    """

    value = True
    try:
        shutil.copy(old_name, new_name)
    except Exception:
        value = False
    return value


def file_move(old_name, new_name):
    """
    Moves a file
    :param old_name: Fully qualified file path to file to be moved.
    :type old_name: str
    :param new_name: Fully qualified path to new file location and name.
    :type new_name: str
    :return: True file moved, otherwise False
    :rtype: bool
    """

    value = True
    try:
        shutil.move(old_name, new_name)
    except Exception:
        value = False
    return value


def move_all_files_from_dir_to_dir(src_dir, dest_dir):
    """
    Moves all files from one directory to another.
    :param src_dir: Source directory path
    :type src_dir: str
    :param dest_dir: Destination directory path
    :type dest_dir: str
    :return: True if moved, otherwise False
    :rtype: bool
    """

    value = True
    try:
        for file_name in os.listdir(src_dir):
            full_file_name = os.path.join(src_dir, file_name)
            if os.path.isfile(full_file_name):
                shutil.move(full_file_name, dest_dir)
    except Exception:
        value = False
    return value

#: file size in KB conversion
FILE_SIZE_IN_KB = 1024
#: file size in MB conversion
FILE_SIZE_IN_MB = 1024 * 1024
#: file size in GB conversion
FILE_SIZE_IN_GB = 1024 * 1024 * 1024


def get_file_size(file_path, unit=FILE_SIZE_IN_MB):
    """
    Get the file size in given units (default is MB)
    :param file_path: Fully qualified file path
    :type file_path: str
    :param unit: the file size unit, defaults to FILE_SIZE_IN_MB
    :type unit: int
    :return: The file size.
    :rtype: float
    """

    # default value if anything goes wrong
    size = -1
    try:
        size = os.path.getsize(file_path)
        # convert units
        size = size / unit
    except:
        pass
    return size


def convert_relative_path_to_full_path(relative_file_path, full_file_path):
    """
    removes '../..' or '../' from relative file path string and replaces it with full path derived path past in sample path.
    - relative path sample: 'C:/temp/../myfile.ext'
    - full file path sample: 'C:/temp/Sample/someOtherFile.ext'
    - returns: 'C:/temp/Sample/myfile.ext'
    :param relative_file_path: String containing relative file path annotation.
    :type relative_file_path: str
    :param full_file_path: A fully qualified file path of which the relative file path is a sub set.
    :type full_file_path: str
    :return: A fully qualified file path.
    :rtype: str
    """

    if r"..\.." in relative_file_path:
        two_up = os.path.abspath(os.path.join(full_file_path, r"..\.."))
        return two_up + relative_file_path[5:]
    elif ".." in relative_file_path:
        one_up = os.path.abspath(os.path.join(full_file_path, ".."))
        return one_up + relative_file_path[2:]
    else:
        return relative_file_path


def get_file_name_without_ext(file_path):
    """
    Returns the file name without the file extension.

    :param file_path: Fully qualified file path to file
    :type file_path: str
    :return: The file name.
    :rtype: str
    """

    file_name = os.path.basename(file_path)
    name_chunks = file_name.split(".")
    name = ""
    if len(name_chunks) > 1:
        name = ".".join(name_chunks[:-1])
    else:
        name = file_name
    return name


def get_file_extension(file_path):
    """
    Returns the file extension of give file name.

    :param file_path: The file name. Can be just the file name or also the fully qualified file path.
    :type file_path: str
    :return: The file extension in format '.extension'
    :rtype: str
    """

    # this will return a tuple of root and extension
    split_tup = os.path.splitext(file_path)
    # extract the file extension
    file_extension = split_tup[1]
    return file_extension


def read_text_file(file_path):
    """
    Reads a text file and returns its content as a single string.

    :param file_path: The fully qualified file path.
    :type file_path: str
    :return: The content of a text file.
    :rtype: str
    """

    file_content = None
    try:
        with open(file_path, "r") as file:
            file_content = file.read()
            file.close()
    except Exception:
        file_content = None
    return file_content


def read_text_file_into_list(file_path):
    """
    Reads a text file and returns its content as a list of string.

    It will return

    - one string per row
    - removed line break characters

    :param file_path: The fully qualified file path.
    :type file_path: str
    :return: The content of a text file. Can be an empty list if an exception occurred during file read.
    :rtype: [str]
    """

    lines = []
    file_content = read_text_file(file_path)
    if file_content != None:
        lines = file_content.splitlines()
    return lines


def get_first_row_in_file(file_path):
    """
    Reads the first line of a text file and returns it as a single string with any leading or trailing white spaces stripped!

    :param file_path: The fully qualified file path.
    :type file_path: str
    :return: The first row of a text file.
    :rtype: str
    """

    row = ""
    try:
        with open(file_path) as f:
            row = f.readline().strip()
    except Exception:
        row = None
    return row


def get_first_row_in_file_no_strip(file_path):
    """
    Reads the first line of a text file and returns it as a single string.

    Note this may contain a new line character at the end! ("\\n")

    :param file_path: The fully qualified file path.
    :type file_path: str
    :return: The first row of a text file.
    :rtype: str
    """

    row = ""
    try:
        with open(file_path) as f:
            row = f.readline()
    except Exception:
        row = None
    return row


def remove_backup_revit_files_from_list(files):
    """
    Takes a list of revit files or full paths and removes any backup files from the list.

    :param files: List of revit files or full paths
    :type files: list
    :return: List of revit files or full paths with backup files removed
    :rtype: list
    """
    files_not_backups = []
    # Exclude any files that are backups: they have .0001.rvt (or similar pattern) in their name
    for rvt_file in files:
        if not is_back_up_file(rvt_file):
            files_not_backups.append(rvt_file)

    return files_not_backups


def is_back_up_file(file_path):
    """
    Checks whether a file is a Revit back up file.

    Backup files are usually in format 'filename.01234.ext'

    Method of checking:

    - splitting file name at every full stop
    - check whether a list with more more then 2 entries came back ?

        - no:
            - not a back up
        - yes:
            - check last list entry whether it is 4 characters in length and can it be convert it into an integer?

                - yes:
                    - backup file
                - no
                    - normal file

    :param file_path: A fully qualified file path.
    :type file_path: str

    :return: True if a back up file, otherwise False.
    :rtype: bool
    """

    is_backup = False
    chunks = file_path.split(".")
    if len(chunks) > 2:
        last_chunk = chunks[len(chunks) - 2]
        try:
            if len(last_chunk) == 4:
                converted_num = int(last_chunk)
                is_backup = True
            else:
                is_backup = False
        except Exception:
            pass
    return is_backup


def remove_null_bytes(file_path, temp_file_path):
    """
    Remove null bytes from a source txt file and write to a temporary file.

    :param file_path: The path to the original txt file.
    :type file_path: str
    :param temp_file_path: The path to the temporary file where cleaned data is written.
    :type temp_file_path: str
    """
    with open(file_path, 'rb') as source_file:
        # Read the entire file and replace null bytes with an empty string
        content = source_file.read().replace(b'\0', b'')

    # Write content to a temporary file
    with open(temp_file_path, 'wb') as temp_file:
        temp_file.write(content)



def is_last_char_newline(file_path):
    """
    Check if the last character in a file is a newline character.

    :param file_path: The fully qualified file path.
    :type file_path: str
    :return: True if the last character is a newline, otherwise False.
    :rtype: bool
    """

    # Function to check if the last character in the file is a newline
    # Check if the file exists
    if not os.path.exists(file_path):
        return False

    with open(file_path, 'rb') as file:
        # Seek to the last byte of the file
        file.seek(-1, os.SEEK_END)
        
        # Read the last byte and check if it's a newline character
        last_char = file.read(1)
        
        return last_char == b'\n'



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Module containing functions relating to family type xml file creation.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Reports:

- all family located in library folder(s)

"""

# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import os
import datetime
import tempfile

from duHast.UI.file_list import get_revit_files
from duHast.Utilities.files_io import get_file_name_without_ext, file_exist
from duHast.Utilities.files_xml import read_xml_file
from duHast.Utilities.directory_io import directory_exists
from duHast.Utilities.Objects.timer import Timer
from duHast.Utilities.Objects.result import Result
from duHast.UI.Objects.ProgressBase import ProgressBase


def get_families_from_directory(directory):
    """
    Get all family files from a directory. (ignores subdirectories)

    :param directory: directory to search for family files
    :type directory: str

    :return: list of family files
    :rtype: [FileItem]
    """

    family_files = get_revit_files(directory, "*.rfa")
    return family_files


def get_xml_files_from_directory(directory):
    """
    Get all xml files from a directory. (ignores subdirectories)

    :param directory: directory to search for xml files
    :type directory: str

    :return: list of xml files
    :rtype: [FileItem]
    """

    xml_files = get_revit_files(directory, "*.xml")
    return xml_files


def check_file_item_exists(instances, name_to_check):
    """
    Check if a file item exists in a list of file items based on the name property.

    :param instances: list of file items
    :type instances: [FileItem]
    :param name_to_check: name to check
    :type name_to_check: str

    :return: True if a file item with the name exists, False otherwise
    :rtype: bool
    """

    return any(instance.name == name_to_check for instance in instances)


def get_families_requiring_update(family_files, xml_files):
    """
    Check if any family files require an xml file update

    :param family_files: the family files to check
    :type family_files: [:class:`FileItem`]
    :param xml_files: the xml files to check against
    :type xml_files: [:class:`FileItem`]

    :return: a list of family files that require updating
    :rtype: list
    """

    return_value = Result()
    family_files_to_update = []
    try:
        # loop over family files
        for fam_file in family_files:

            # and find a matching xml file
            xml_file = os.path.splitext(fam_file.name)[0] + ".xml"

            # get the catalogue file too
            catalogue_file = os.path.splitext(fam_file.name)[0] + ".txt"

            # check if the xml file does exist
            if check_file_item_exists(xml_files, xml_file):

                # get time stamp of files for varies file types (.rfa, .xml, .txt) relating to families
                # .txt catalogue file
                # .xml part atom export file
                # .rfa family file
                xml_file_time_stamp = os.path.getmtime(xml_file)
                rfa_file_time_stamp = os.path.getmtime(fam_file.name)
                txt_file_time_stamp = (
                    None
                    if not (file_exist(catalogue_file))
                    else os.path.getmtime(catalogue_file)
                )

                # convert to something human readable
                xml_formatted_time = datetime.datetime.fromtimestamp(
                    xml_file_time_stamp
                )
                rfa_formatted_time = datetime.datetime.fromtimestamp(
                    rfa_file_time_stamp
                )
                txt_formatted_time = (
                    None
                    if txt_file_time_stamp is None
                    else datetime.datetime.fromtimestamp(txt_file_time_stamp)
                )

                # if the xml file is older than the family file
                if xml_file_time_stamp < rfa_file_time_stamp:

                    return_value.append_message(
                        "...xml file is older than family file, needs updating. [xml: {} vs rfa: {}]".format(
                            xml_formatted_time, rfa_formatted_time
                        )
                    )
                    # family xml file is older than family file, needs updating
                    family_files_to_update.append(fam_file.name)
                elif (
                    txt_file_time_stamp is not None
                    and xml_file_time_stamp < txt_file_time_stamp
                ):
                    return_value.append_message(
                        "...xml file is older than catalogue file, needs updating. [xml: {} vs txt: {}]".format(
                            xml_formatted_time, txt_formatted_time
                        )
                    )
                    # family xml file is older than family file, needs updating
                    family_files_to_update.append(fam_file.name)
                else:
                    # family xml file is up to date
                    pass

            else:
                # no matching xml file found
                family_files_to_update.append(fam_file.name)

        # save data to be returned
        return_value.result = family_files_to_update
    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to check families requiring update with exception: {}".format(e),
        )

    return return_value


def write_data_to_xml_file(application, family_path, xml_path):
    """
    Write the family type data to an XML file.

    :param application: The Revit application object.
    :type application: Autodesk.Revit.ApplicationServices.Application
    :param family_path: The path of the family file.
    :type family_path: str
    :param xml_path: The path of the XML file.
    :type xml_path: str

    :return: A result object with .status True if successful.
    :rtype: Result
    """

    return_value = Result()
    try:

        # Save XML file to temporary location
        # this is a method of the application object and does not require the family to be open...
        application.ExtractPartAtomFromFamilyFile(family_path, xml_path)
        return_value.update_sep(True, "Wrote data to XML file.")
    except Exception as e:
        return_value.update_sep(False, "Failed to write XML data: {}".format(e))

    return return_value


def create_xml_file(revit_application, family_file):
    """
    Create an xml file for a given family file.

    :param revit_application: revit application object
    :type revit_application: Application
    :param family_file: family file to create xml file for
    :type family_file: str
    """

    return_value = Result()
    try:
        xml_file = os.path.splitext(family_file)[0] + ".xml"
        result_write = write_data_to_xml_file(revit_application, family_file, xml_file)
        return_value.update_sep(result_write)
    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to create xml file for family: {} with exception: {}".format(
                family_file, e
            ),
        )
    return return_value


def write_data_to_temp_xml_file_and_read_it_back(an_action_to_write_xml_data):
    """
    Write the data to a temp XML file and read it back.

    :param an_action_to_write_xml_data: The action to write the XML data.
    :type an_action_to_write_xml_data: function returning a Result object

    :return:
        Result class instance.

        - result.status: True if data was written and read back successfully, False otherwise.
        - result.message will contain the log data.
        - result.result will be a XML document object.

        On exception

        - Reload.status (bool) will be False
        - Reload.message will contain the exception message
        - Reload.result will be an empty list.

    :rtype: :class:`.Result`
    """

    return_value = Result()

    # Create a temporary file
    with tempfile.NamedTemporaryFile(delete=False, suffix=".xml") as temp_file:
        temp_path_xml = temp_file.name

    try:

        # Write the data to the file
        write_result = an_action_to_write_xml_data(temp_path_xml)
        # update the return value
        return_value.update(write_result)

        # Check if the write was successful
        if return_value.status is False:
            return return_value

        # Read the data back from the file
        read_result = read_xml_file(temp_path_xml)

        # update the return value message and status
        # for some reasons this adds a XMLDeclaration object to the result field...not sure why
        return_value.update(read_result)

        # overwrite result field with the actual XML document
        return_value.result = read_result.result

    finally:
        # Delete the temporary file
        if os.path.exists(temp_path_xml):
            os.remove(temp_path_xml)

    return return_value


def write_data_to_xml_file_and_read_it_back(an_action_to_write_xml_data, xml_file_path):
    """
    Write the data to an XML file and read it back.

    :param an_action_to_write_xml_data: The action to write the XML data.
    :type an_action_to_write_xml_data: function
    :param xml_file_path: The path of the XML file.
    :type xml_file_path: str

    :return:
        Result class instance.

        - result.status: True if data was written and read back successfully, False otherwise.
        - result.message will contain log data
        - result.result will be a XML document object.

        On exception

        - Reload.status (bool) will be False
        - Reload.message will contain the exception message
        - Reload.result will be an empty list.

    :rtype: :class:`.Result`
    """

    return_value = Result()

    try:

        # Write the data to the file
        write_result = an_action_to_write_xml_data(xml_file_path)

        # update the return value
        return_value.update(write_result)

        # Check if the write was successful
        if return_value.status is False:
            return return_value

        # Read the data back from the file
        read_result = read_xml_file(xml_file_path)

        # update the return value message and status
        # for some reasons this adds a XMLDeclaration object to the result field...not sure why
        return_value.update(read_result)
        # overwrite result field with the actual XML document
        return_value.result = read_result.result

    except Exception as e:
        return_value.update_sep(False, "{}".format(e))
    return return_value


def create_family_xml_files(
    revit_application, process_directories, progress_callback=None
):
    """
    Create xml files for all families in a list of directories using Revit API PartAtomExport function.

    :param revit_application: revit application object
    :type revit_application: Application
    :param process_directories: list of directories to process
    :type process_directories: [str]
    :param progress_callback: progress callback object
    :type progress_callback: ProgressBase

    :return: Result object
    :rtype: Result
    """

    return_value = Result()

    # check callback class
    if progress_callback and isinstance(progress_callback, ProgressBase) == False:
        raise TypeError(
            "progress_callback needs to be inherited from ProgressBase. Got : {} instead.".format(
                type(progress_callback)
            )
        )

    # check if directories are valid
    if isinstance(process_directories, list) == False:
        raise TypeError(
            "progress_directories needs to be a list. Got: {} instead.".format(
                type(process_directories)
            )
        )

    # set up a timer
    t = Timer()
    t.start()

    try:

        # loop through directories
        for directory in process_directories:

            # check if directory exists
            if directory_exists(directory) == False:
                return_value.update_sep(
                    False, "Directory does not exist: {} Skipping it.".format(directory)
                )
                continue

            # get any revit families from the directory
            return_value.append_message("Processing directory: {}".format(directory))

            families = get_families_from_directory(directory=directory)
            return_value.append_message("Families found: {}".format(len(families)))

            if len(families) == 0:
                return_value.append_message(
                    "No families found in directory: {}".format(directory)
                )
                continue

            # get any xml files from the directory
            xml_files = get_xml_files_from_directory(directory=directory)
            return_value.append_message("XML files found: {}".format(len(xml_files)))

            # check if any families require updating ( a new xml file is required or an existing one is out of date)
            families_to_update_result = get_families_requiring_update(
                families, xml_files
            )
            if families_to_update_result.status == False:
                return_value.update_sep(False, families_to_update_result.message)
                return_value.append_message(t.stop())
                return return_value

            families_to_update = families_to_update_result.result
            return_value.append_message(
                "Families requiring update: {}".format(len(families_to_update))
            )

            # check if any families require updating
            if families_to_update:

                # set up progress
                fam_counter = 1
                max_fam = len(families_to_update)

                # iterate through families
                for family in families_to_update:

                    # get the family name
                    fam_name = get_file_name_without_ext(family)

                    # update progress
                    if progress_callback:
                        progress_callback.update(fam_counter, max_fam, fam_name)

                    create_xml_file(revit_application, family)
                    return_value.append_message(
                        "Created xml file for family: {}".format(fam_name)
                    )

                    # update counter
                    fam_counter += 1

                    # check for user cancel
                    if progress_callback != None:
                        if progress_callback.is_cancelled():
                            return_value.append_message("User cancelled!")
                            break

        # stop the timer
        return_value.append_message(
            "Successfully created family atom exports: {}".format(t.stop())
        )

    except Exception as e:
        return_value.update_sep(
            False, "Failed to create family atom exports with exception: {}".format(e)
        )
        if t.is_running():
            return_value.append_message(t.stop())

    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Model health report functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Model health report metrics can either be displayed in a family where each parameter is assigned to a metric 
and or data can be exported to text files which can be used to visualize key metrics over time.

"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


import clr
import os
import System
from csv import QUOTE_MINIMAL

clr.AddReference("System.Core")
from System import Linq

clr.ImportExtensions(Linq)

from System.Collections.Generic import List
from collections import namedtuple

from Autodesk.Revit.DB import (
    ParameterValueProvider,
    BuiltInParameter,
    ElementId,
    ElementParameterFilter,
    FamilyInstance,
    FilteredElementCollector,
    FilterStringEquals,
    FilterStringRule,
)

from duHast.Utilities.Objects import result as res
from duHast.Revit.ModelHealth.Reporting import report_file_names as rFns
from duHast.Utilities.files_io import get_file_name_without_ext
from duHast.Utilities.date_stamps import (
    get_file_date_stamp,
    get_date_stamp,
    FILE_DATE_STAMP_YYYYMMDD_SPACE,
    TIME_STAMP_HHMMSEC_COLON,
)

from duHast.Revit.ModelHealth.Reporting.Properties.view import (
    get_number_of_sheets,
    get_number_of_unused_view_filters,
    get_number_of_unplaced_views,
    get_number_of_unused_view_templates,
    get_number_of_view_filters,
    get_number_of_view_templates,
    get_number_of_views,
)

from duHast.Revit.ModelHealth.Reporting.Properties.rooms import (
    get_number_of_not_enclosed_rooms,
    get_number_of_redundant_rooms,
    get_number_of_rooms,
    get_number_of_unplaced_rooms,
)

from duHast.Revit.ModelHealth.Reporting.Properties.constants import (
    FAILED_TO_RETRIEVE_VALUE,
    MODEL_HEALTH_TRACKER_FAMILY,
)

from duHast.Revit.ModelHealth.Reporting.Properties.design_set_options import (
    get_number_of_design_options,
    get_number_of_design_sets,
)

from duHast.Revit.ModelHealth.Reporting.Properties.line_styles_types import (
    get_number_of_line_styles,
    get_number_of_fill_patterns,
    get_number_of_line_patterns,
)

from duHast.Revit.ModelHealth.Reporting.Properties.links_cad import (
    get_number_of_cad_imports,
    get_number_of_cad_links_to_model,
    get_number_of_cad_links_to_view,
)

from duHast.Revit.ModelHealth.Reporting.Properties.groups import (
    get_number_of_detail_groups,
    get_number_of_model_groups,
    get_number_of_unplaced_detail_groups,
    get_number_of_unplaced_model_groups,
)

from duHast.Revit.ModelHealth.Reporting.Properties.annotations import (
    get_number_of_dimension_types,
    get_number_of_unused_dimension_types,
    get_number_of_text_types,
    get_number_of_unused_text_types,
    get_number_of_arrow_head_types,
    get_number_of_unused_arrow_head_types,
)

from duHast.Revit.ModelHealth.Reporting.Properties.links_images import (
    get_number_of_image_imports,
    get_number_of_image_links,
)

from duHast.Revit.ModelHealth.Reporting.Properties.general import (
    get_number_of_warnings,
    get_current_date,
    get_workset_number,
    get_current_file_size,
)

from duHast.Revit.ModelHealth.Reporting.Properties.families import (
    get_number_of_families,
    get_number_of_in_place_families,
)

from duHast.Revit.ModelHealth.Reporting.Properties.detail_items import (
    get_number_of_filled_regions,
)


from duHast.Revit.Common.parameter_set_utils import set_parameter_value
from duHast.Utilities.files_csv import write_report_data_as_csv
from duHast.Revit.Common.revit_version import get_revit_version_number


def _cast_parameter_value(p_value):
    """
    Check if parameter is of type string ( currently the date only)
    and only cast to string if not...

    :param p_value: The parameter value
    :type p_value: unknown
    :return: The parameter value as a string
    :rtype: str
    """

    new_para_value = ""
    if p_value.GetType() != System.String:
        new_para_value = str(p_value)
    else:
        new_para_value = p_value
    return new_para_value


def get_instances_of_model_health(doc):
    """
    Gets all instances of the model health tracker family in a model.

    Built in parameter containing family name when filtering familyInstance elements:
    BuiltInParameter.ELEM_FAMILY_PARAM
    This is a faster filter in terms of performance then LINQ query refer to:
    https://jeremytammik.github.io/tbc/a/1382_filter_shortcuts.html

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A list containing all model health tracker families in the model.
    :rtype: list of Autodesk.Revit.DB.FamilyInstance
    """

    provider = ParameterValueProvider(ElementId(BuiltInParameter.ELEM_FAMILY_PARAM))
    evaluator = FilterStringEquals()

    revit_version = get_revit_version_number(doc=doc)
    # define rule with a placeholder
    rule = None
    # ge the rule depending on Revit version
    if revit_version <= 2022:
        # use case sensitive flag
        rule = FilterStringRule(provider, evaluator, MODEL_HEALTH_TRACKER_FAMILY, True)
    else:
        # case sensitive flag has been removed in Revit 2023 onwards
        rule = FilterStringRule(provider, evaluator, MODEL_HEALTH_TRACKER_FAMILY)

    filter = ElementParameterFilter(rule)
    return (
        FilteredElementCollector(doc)
        .OfClass(FamilyInstance)
        .WherePasses(filter)
        .ToList()
    )


def get_parameters_of_instance(fam_instance, doc):
    """
    Updates parameter values of model tracker family instance.

    :param fam_instance: An instance of the model health tracker family.
    :type fam_instance: Autodesk.Revit.DB.FamilyInstance
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return:
        Result class instance.

        - .status True if all parameters where found on the family and got updated successfully or no update at all was required. Otherwise False.
        - .message will be 'Failed to get value for'

    :rtype: :class:`.Result`
    """

    result_value = res.Result()
    flag_update = False
    for p in fam_instance.GetOrderedParameters():
        # check if parameter is read only
        if p.IsReadOnly == False:
            # check an action to update this parameter value exists
            if PARAM_ACTIONS.ContainsKey(p.Definition.Name):
                try:
                    parameter_value = PARAM_ACTIONS[p.Definition.Name].get_data(doc)
                    if parameter_value != FAILED_TO_RETRIEVE_VALUE:
                        flag = set_parameter_value(
                            p, _cast_parameter_value(parameter_value), doc
                        )
                        result_value.update(flag)
                        flag_update = True
                    else:
                        result_value.update_sep(
                            False, "Failed to get value for " + p.Definition.Name
                        )
                except Exception as e:
                    result_value.update_sep(
                        False,
                        "Failed to update {} with exception: {}".format(
                            p.Definition.Name, e
                        ),
                    )
    if flag_update == False:
        result_value.append_message( "No family parameters where updated")
        result_value.status = True
    return result_value


# ----------------------------------------------
# main
# ----------------------------------------------

# set up a named tuple to store data in it
health_data_action = namedtuple("healthDataAction", "get_data report_file_name")

#: List of actions reporting model health metrics and their associated parameter name
PARAM_ACTIONS = {
    "ValueWorksets": health_data_action(
        get_workset_number, rFns.PARAM_ACTIONS_FILENAME_NO_OF_WORKSETS
    ),
    "ValueFileSize": health_data_action(
        get_current_file_size, rFns.PARAM_ACTIONS_FILENAME_FILE_SIZE
    ),
    "ValueWarnings": health_data_action(
        get_number_of_warnings, rFns.PARAM_ACTIONS_FILENAME_NO_OF_WARNINGS
    ),
    "ValueDesignSets": health_data_action(
        get_number_of_design_sets, rFns.PARAM_ACTIONS_FILENAME_NO_OF_DESIGN_SETS
    ),
    "ValueDesignOptions": health_data_action(
        get_number_of_design_options, rFns.PARAM_ACTIONS_FILENAME_NO_OF_DESIGN_OPTIONS
    ),
    "ValueSheets": health_data_action(
        get_number_of_sheets, rFns.PARAM_ACTIONS_FILENAME_NO_OF_SHEETS
    ),
    "ValueViews": health_data_action(
        get_number_of_views, rFns.PARAM_ACTIONS_FILENAME_NO_OF_VIEWS
    ),
    "ValueViewsNotPlaced": health_data_action(
        get_number_of_unplaced_views, rFns.PARAM_ACTIONS_FILENAME_NO_OF_VIEWS_NOT_PLACED
    ),
    "ValueViewFilters": health_data_action(
        get_number_of_view_filters, rFns.PARAM_ACTIONS_FILENAME_NO_OF_VIEW_FILTERS
    ),
    "ValueViewFiltersUnused": health_data_action(
        get_number_of_unused_view_filters,
        rFns.PARAM_ACTIONS_FILENAME_NO_OF_VIEW_FILTERS_UNUSED,
    ),
    "ValueViewTemplates": health_data_action(
        get_number_of_view_templates, rFns.PARAM_ACTIONS_FILENAME_NO_OF_VIEW_TEMPLATES
    ),
    "ValueViewTemplatesUnused": health_data_action(
        get_number_of_unused_view_templates,
        rFns.PARAM_ACTIONS_FILENAME_NO_OF_VIEW_TEMPLATES_UNUSED,
    ),
    "ValueLineStyles": health_data_action(
        get_number_of_line_styles, rFns.PARAM_ACTIONS_FILENAME_NO_OF_LINE_STYLES
    ),
    "ValueLinePatterns": health_data_action(
        get_number_of_line_patterns, rFns.PARAM_ACTIONS_FILENAME_NO_OF_LINE_PATTERNS
    ),
    "ValueFillPatterns": health_data_action(
        get_number_of_fill_patterns, rFns.PARAM_ACTIONS_FILENAME_NO_OF_FILL_PATTERNS
    ),
    "ValueCADImports": health_data_action(
        get_number_of_cad_imports, rFns.PARAM_ACTIONS_FILENAME_NO_OF_CAD_IMPORTS
    ),
    "ValueCADLinksToModel": health_data_action(
        get_number_of_cad_links_to_model,
        rFns.PARAM_ACTIONS_FILENAME_NO_OF_CAD_LINKS_MODEL,
    ),
    "ValueCADLinksToView": health_data_action(
        get_number_of_cad_links_to_view,
        rFns.PARAM_ACTIONS_FILENAME_NO_OF_CAD_LINKS_VIEW,
    ),
    "ValueImageImports": health_data_action(
        get_number_of_image_imports, rFns.PARAM_ACTIONS_FILENAME_NO_OF_IMAGE_IMPORTS
    ),
    "ValueImageLinks": health_data_action(
        get_number_of_image_links, rFns.PARAM_ACTIONS_FILENAME_NO_OF_IMAGE_LINKS
    ),
    "ValueFamilies": health_data_action(
        get_number_of_families, rFns.PARAM_ACTIONS_FILENAME_NO_OF_FAMILIES
    ),
    "ValueFamiliesInPlace": health_data_action(
        get_number_of_in_place_families,
        rFns.PARAM_ACTIONS_FILENAME_NO_OF_FAMILIES_IN_PLACE,
    ),
    "ValueModelGroups": health_data_action(
        get_number_of_model_groups, rFns.PARAM_ACTIONS_FILENAME_NO_OF_MODEL_GROUPS
    ),
    "ValueModelGroupsUnplaced": health_data_action(
        get_number_of_unplaced_model_groups,
        rFns.PARAM_ACTIONS_FILENAME_NO_OF_MODEL_GROUPS_UNPLACED,
    ),
    "ValueDetailGroups": health_data_action(
        get_number_of_detail_groups, rFns.PARAM_ACTIONS_FILENAME_NO_OF_DETAIL_GROUPS
    ),
    "ValueDetailGroupsUnplaced": health_data_action(
        get_number_of_unplaced_detail_groups,
        rFns.PARAM_ACTIONS_FILENAME_NO_OF_DETAIL_GROUPS_UNPLACED,
    ),
    "ValueRooms": health_data_action(
        get_number_of_rooms, rFns.PARAM_ACTIONS_FILENAME_NO_OF_ROOMS
    ),
    "ValueRoomsUnplaced": health_data_action(
        get_number_of_unplaced_rooms, rFns.PARAM_ACTIONS_FILENAME_NO_OF_ROOMS_UNPLACED
    ),
    "ValueRoomsNotEnclosed": health_data_action(
        get_number_of_not_enclosed_rooms,
        rFns.PARAM_ACTIONS_FILENAME_NO_OF_ROOMS_UNENCLOSED,
    ),
    "ValueRoomsRedundant": health_data_action(
        get_number_of_redundant_rooms, rFns.PARAM_ACTIONS_FILENAME_NO_OF_ROOMS_REDUNDANT
    ),
    "ValueFilledRegions": health_data_action(
        get_number_of_filled_regions, rFns.PARAM_ACTIONS_FILENAME_NO_OF_FILLED_REGIONS
    ),
    "ValueDimensionStyles": health_data_action(
        get_number_of_dimension_types,
        rFns.PARAM_ACTIONS_FILENAME_NO_OF_DIMENSION_STYLES,
    ),
    "ValueDimensionStylesUnused": health_data_action(
        get_number_of_unused_dimension_types,
        rFns.PARAM_ACTIONS_FILENAME_NO_OF_DIMENSION_STYLES_UNUSED,
    ),
    "ValueTextStyles": health_data_action(
        get_number_of_text_types, rFns.PARAM_ACTIONS_FILENAME_NO_OF_TEXT_STYLES
    ),
    "ValueTextStylesUnused": health_data_action(
        get_number_of_unused_text_types,
        rFns.PARAM_ACTIONS_FILENAME_NO_OF_TEXT_STYLES_UNUSED,
    ),
    "ValueArrowHeadStyles": health_data_action(
        get_number_of_arrow_head_types,
        rFns.PARAM_ACTIONS_FILENAME_NO_OF_ARROW_HEAD_STYLES,
    ),
    "ValueArrowHeadStylesUnused": health_data_action(
        get_number_of_unused_arrow_head_types,
        rFns.PARAM_ACTIONS_FILENAME_NO_OF_ARROW_HEAD_STYLES_UNUSED,
    ),
    "ValueDateLastUpdated": health_data_action(
        get_current_date, rFns.PARAM_ACTIONS_FILENAME_DATE_LAST_UPDATED
    ),
}


def update_model_health_tracer_family(doc, revit_file_path):
    """
    Updates instances of model health tracker family in project.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param revit_file_path: Fully qualified revit model file path.
    :type revit_file_path: str

    :return:
        Result class instance.

        - .status True if all model key health metric where updated successfully. Otherwise False.
        - .message will be listing each parameter update: old value to new value

    :rtype: :class:`.Result`
    """

    revit_file_name = get_file_name_without_ext(revit_file_path)
    result_value = res.Result()
    instances = get_instances_of_model_health(doc)
    if len(instances) > 0:
        for instance in instances:
            try:
                update_flag = get_parameters_of_instance(instance, doc)
                result_value.update(update_flag)
            except Exception as e:
                result_value.update_sep(False, "{}".format(e))
    else:
        result_value.update_sep(
            False,
            "Family to update: {} was not found in model: {}".format(
                MODEL_HEALTH_TRACKER_FAMILY, revit_file_name
            ),
        )
    return result_value


# doc   current document
# revitFilePath     path of the current document
def write_model_health_report(doc, revit_file_path, output_directory):
    """
    Write out health tracker data to file.

    Each value gets written to a separate file. The file name is made up of time stamp and the revit file name.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param revit_file_path: Fully qualified revit model file path.
    :type revit_file_path: str
    :param output_directory: The directory path of where to write the data to.
    :type output_directory: str

    :return:
        Result class instance.

        - .status True if data was written to files successfully. Otherwise False.
        - .message will be contain data file path for each file.

    :rtype: :class:`.Result`
    """

    revit_file_name = get_file_name_without_ext(revit_file_path)
    result_value = res.Result()
    # get values and write them out
    for key, value in PARAM_ACTIONS.items():
        # parameter value get may throw an exception
        # record -1 in that case!
        try:
            parameter_value = PARAM_ACTIONS[key].get_data(doc)
        except Exception as e:
            parameter_value = FAILED_TO_RETRIEVE_VALUE
        file_name = (
            get_file_date_stamp()
            + revit_file_name
            + PARAM_ACTIONS[key].report_file_name
            + ".temp"
        )
        res_export = res.Result()
        try:
            # write data to file
            write_result = write_report_data_as_csv(
                file_name= os.path.join(output_directory, file_name),
                header = rFns.LOG_FILE_HEADER,
                data = [
                    [
                        revit_file_name,
                        key,
                        get_date_stamp(FILE_DATE_STAMP_YYYYMMDD_SPACE),
                        get_date_stamp(TIME_STAMP_HHMMSEC_COLON),
                        _cast_parameter_value(parameter_value),
                    ]
                ],
                enforce_ascii=True,
                encoding="utf-8",
                bom=None,
                quoting=QUOTE_MINIMAL,
            )
            if write_result.status == False:
                raise ValueError(write_result.message)
            
            res_export.update_sep(True, "Exported: {}".format(key))
        except Exception as e:
            res_export.update_sep(
                False, "Export failed: {} with exception: {}".format(key, e)
            )
        result_value.update(res_export)
    return result_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit view schedules of category sheet export to file. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import os

from duHast.Revit.Views.schedules import  get_all_sheet_schedules
from duHast.Revit.Views.schedules_fields import schedule_contains_sheet_number_field, get_field_column_index_from_schedule_by_parameter_id, SHEET_NUMBER_PARAMETER_ID 
from duHast.Revit.Views.schedules_export import export_schedule_to_file

from duHast.Utilities.Objects.result import Result
from duHast.Utilities.directory_io import directory_exists, create_temp_directory, directory_delete
from duHast.Utilities.files_get import get_files_single_directory
from duHast.Utilities.files_io import get_file_name_without_ext
from duHast.Utilities.file_base_read_net import read_from_delimited_text_file


def filter_schedules_by_sheet_number_field(schedules):
    """
    Filters the given schedules to only those that contain a visible sheet number field.
    
    :param schedules: The list of schedules to filter.
    :type schedules: list[Autodesk.Revit.DB.ViewSchedule]

    :return: A list of schedules that contain the sheet number field.
    :rtype: list[Autodesk.Revit.DB.ViewSchedule]
    """

    filtered_schedules = []

    for schedule in schedules:
        # check if the schedule contains the sheet number field
        if schedule_contains_sheet_number_field(schedule = schedule, ignore_hidden_field=True):
            filtered_schedules.append(schedule)
    
    return filtered_schedules


def export_all_sheet_schedules_to_file( doc, directory_path, export_if_number_is_hidden = False):
    """
    Exports all schedules that are placed on sheets to file.
    
    :param directory_path: The directory path to export the schedules to.
    :type directory_path: str
    :param export_if_number_is_hidden: Whether to export schedules even if the number is hidden.
    :type export_if_number_is_hidden: bool, optional
    
    :return: Result object containing success status and messages.
    :rtype: duHast.Utilities.Objects.result.Result
    """
    
    return_value = Result()

    try:

        # check if the directory exists
        if directory_exists(directory_path) is False:
            return_value.update_sep(False, "Directory does not exist: {directory_path}".format(directory_path=directory_path))
            return return_value

        # get all sheet schedules in doc
        all_schedules = get_all_sheet_schedules(doc)

        # check if there are any schedules to export
        if len(all_schedules) == 0:
            return_value.update_sep(False, "No sheet schedules in document.")
            return return_value
        
        # check if schedules need to be filtered ( export only if sheet number field is not hidden)
        if export_if_number_is_hidden is False:
            # filter schedules to only those that have the sheet number field visible
            all_schedules = filter_schedules_by_sheet_number_field(all_schedules)

        # iterate through all schedules and export them to file
        for schedule in all_schedules:
            # get the file name from the schedule name
            file_name = schedule.Name + ".csv"
            file_path = os.path.join(directory_path, file_name)
            
            # export the schedule to file
            export_result = export_schedule_to_file(schedule, file_path)
            
            return_value.update(export_result)

        return return_value
    
    except Exception as e:
        return_value.update_sep(False, "Failed to export schedules to file: {e}".format(e=e))
        return return_value


def export_all_sheet_schedules_and_read_data_back(doc, export_if_number_is_hidden=False):
    """
    Exports all sheet schedules to file and reads the data back into a dictionary.

    :param export_if_number_is_hidden: Whether to export schedules even if the number is hidden.
    :type export_if_number_is_hidden: bool, optional

    :return: Result object containing success status, messages, and data.
    :rtype: duHast.Utilities.Objects.result.Result
    """

    return_value = Result()

    try:
        # set up a temporary directory to export the schedules to
        directory_path = create_temp_directory()
        
        # export all sheet schedules to file
        export_result = export_all_sheet_schedules_to_file(doc=doc, directory_path=directory_path, export_if_number_is_hidden=export_if_number_is_hidden)
        
        return_value.update(export_result)

        if return_value.status is False:
            return return_value

        # get all files in temp directory
        if directory_exists(directory_path) is False:
            return_value.update_sep(False, "Temporary directory does not exist: {directory_path}".format(directory_path=directory_path))
            return return_value
        
        # get all files in the directory with the specified filters
        files_found = get_files_single_directory(directory_path, file_prefix="", file_suffix="", file_extension=".csv")

        # check if any files were found
        if len(files_found) == 0:
            return_value.update_sep(False, "No files found in temporary directory: {directory_path}".format(directory_path=directory_path))
            
            # delete the temp directory
            delete_flag = directory_delete(directory_path)
            return_value.append_message("Deleted temporary directory: {directory_path} with status {flag}".format(directory_path=directory_path, flag=delete_flag))

            return return_value
        
        # read the data back into a dictionary
        data = {}

        for file_path in files_found:
            # attempt to read the file
            file_data_result = read_from_delimited_text_file(file_path)

            # check if the file was read successfully
            if file_data_result.status is False:
                return_value.update_sep(False, "Failed to read data from file: {file_path}".format(file_path=file_path))
                continue

            else:
                # log the success message
                return_value.append_message("Successfully read data from file: {file_path}".format(file_path=file_path))
                # store the data in the dictionary with the file name as the key
                file_name = get_file_name_without_ext(file_path)
                data[file_name] = file_data_result.result

        # delete the temporary directory
        delete_flag = directory_delete(directory_path)
        return_value.append_message("Deleted temporary directory: {directory_path} with status {flag}".format(directory_path=directory_path, flag=delete_flag))

        # return the data read
        return_value.result.append(data)
        return return_value
    
    except Exception as e:
        return_value.update_sep(False, "Failed to export and read data back: {e}".format(e=e))
        return return_value


def get_sheet_number_index_to_schedule_mapper(doc):
    """
    Creates a mapping of sheet number column index to their corresponding schedules in the document.

    :param doc: The Revit document containing the schedules.
    :type doc: Autodesk.Revit.DB.Document

    :return: A dictionary mapping sheet numbers to their corresponding schedules.
    :rtype: dict[str, Autodesk.Revit.DB.ViewSchedule]
    """

    return_value = Result()

    try:
        # get all sheet schedules in doc
        all_schedules = get_all_sheet_schedules(doc)

        # filter schedules to only those that have the sheet number field visible
        all_schedules = filter_schedules_by_sheet_number_field(all_schedules)

        # create a mapping of sheet numbers to schedules
        sheet_number_to_schedule_map = {}

        for schedule in all_schedules:
           column_index = get_field_column_index_from_schedule_by_parameter_id(schedule, SHEET_NUMBER_PARAMETER_ID )
           sheet_number_to_schedule_map[schedule.Name] = column_index
        
        return_value.result.append(sheet_number_to_schedule_map)
        return return_value
    
    except Exception as e:
        return_value.update_sep(False, "Failed to create sheet number column index to schedule mapper: {e}".format(e=e))
        return return_value